# sage_setup: distribution = sagemath-polyhedra
"""
Functions for plotting polyhedra
"""

########################################################################
#       Copyright (C) 2008 Marshall Hampton <hamptonio@gmail.com>
#       Copyright (C) 2011 Volker Braun <vbraun.name@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  https://www.gnu.org/licenses/
########################################################################

from math import pi

from sage.misc.lazy_import import lazy_import
from sage.structure.sage_object import SageObject
from sage.modules.free_module_element import vector
from sage.matrix.constructor import matrix, identity_matrix
from sage.matrix.special import diagonal_matrix
from sage.misc.functional import norm
from sage.misc.latex import LatexExpr
from sage.structure.sequence import Sequence

lazy_import("sage.plot.all", ["Graphics", "point2d", "line2d", "arrow", "polygon2d", "rainbow"])
lazy_import("sage.plot.plot3d.all", ["point3d", "line3d", "arrow3d", "polygons3d"])
lazy_import("sage.plot.plot3d.transform", "rotate_arbitrary")
lazy_import("sage.plot.plot3d.texture", "Texture")


#############################################################


def cyclic_sort_vertices_2d(Vlist):
    """
    Return the vertices/rays in cyclic order if possible.

    .. NOTE::

        This works if and only if each vertex/ray is adjacent to exactly
        two others. For example, any 2-dimensional polyhedron satisfies
        this.

    See
    :meth:`~sage.geometry.polyhedron.base.Polyhedron_base.vertex_adjacency_matrix`
    for a discussion of "adjacent".

    EXAMPLES::

        sage: from sage.geometry.polyhedron.plot import cyclic_sort_vertices_2d
        sage: square = Polyhedron([[1,0],[-1,0],[0,1],[0,-1]])
        sage: vertices = [v for v in square.vertex_generator()]
        sage: vertices
        [A vertex at (-1, 0),
         A vertex at (0, -1),
         A vertex at (0, 1),
         A vertex at (1, 0)]
        sage: cyclic_sort_vertices_2d(vertices)
        [A vertex at (1, 0),
         A vertex at (0, -1),
         A vertex at (-1, 0),
         A vertex at (0, 1)]

    Rays are allowed, too::

        sage: P = Polyhedron(vertices=[(0, 1), (1, 0), (2, 0), (3, 0), (4, 1)], rays=[(0,1)])
        sage: P.adjacency_matrix()
        [0 1 0 1 0]
        [1 0 1 0 0]
        [0 1 0 0 1]
        [1 0 0 0 1]
        [0 0 1 1 0]
        sage: cyclic_sort_vertices_2d(P.Vrepresentation())
        [A vertex at (3, 0),
         A vertex at (1, 0),
         A vertex at (0, 1),
         A ray in the direction (0, 1),
         A vertex at (4, 1)]

        sage: P = Polyhedron(vertices=[(0, 1), (1, 0), (2, 0), (3, 0), (4, 1)], rays=[(0,1), (1,1)])
        sage: P.adjacency_matrix()
        [0 1 0 0 0]
        [1 0 1 0 0]
        [0 1 0 0 1]
        [0 0 0 0 1]
        [0 0 1 1 0]
        sage: cyclic_sort_vertices_2d(P.Vrepresentation())
        [A ray in the direction (1, 1),
         A vertex at (3, 0),
         A vertex at (1, 0),
         A vertex at (0, 1),
         A ray in the direction (0, 1)]

        sage: P = Polyhedron(vertices=[(1,2)], rays=[(0,1)], lines=[(1,0)])
        sage: P.adjacency_matrix()
        [0 0 1]
        [0 0 0]
        [1 0 0]
        sage: cyclic_sort_vertices_2d(P.Vrepresentation())
        [A vertex at (0, 2),
         A line in the direction (1, 0),
         A ray in the direction (0, 1)]
    """
    if not Vlist:
        return Vlist
    Vlist = list(Vlist)
    result = []
    adjacency_matrix = Vlist[0].polyhedron().vertex_adjacency_matrix()

    # Any object in Vlist has 0,1, or 2 adjacencies. Break into connected chains:
    chain = [Vlist.pop()]
    while Vlist:
        first_index = chain[0].index()
        last_index = chain[-1].index()
        for v in Vlist:
            v_index = v.index()
            if adjacency_matrix[last_index, v_index] == 1:
                chain = chain + [v]
                Vlist.remove(v)
                break
            if adjacency_matrix[first_index, v.index()] == 1:
                chain = [v] + chain
                Vlist.remove(v)
                break
        else:
            result += chain
            chain = [Vlist.pop()]
    result += chain
    return result

#########################################################################


def projection_func_identity(x):
    """
    The identity projection.

    EXAMPLES::

        sage: from sage.geometry.polyhedron.plot import projection_func_identity
        sage: projection_func_identity((1,2,3))
        [1, 2, 3]
    """
    return list(x)


class ProjectionFuncStereographic:
    """
    The stereographic (or perspective) projection onto a codimension-1 linear
    subspace with respect to a sphere centered at the origin.

    EXAMPLES::

        sage: from sage.geometry.polyhedron.plot import ProjectionFuncStereographic
        sage: cube = polytopes.hypercube(3).vertices()
        sage: proj = ProjectionFuncStereographic([1.2, 3.4, 5.6])
        sage: ppoints = [proj(vector(x)) for x in cube]
        sage: ppoints[5]
        (-0.0918273..., -0.036375...)
    """
    def __init__(self, projection_point):
        """
        Create a stereographic projection function.

        INPUT:

        - ``projection_point`` -- list of coordinates in the
          appropriate dimension, which is the point projected from

        EXAMPLES::

            sage: from sage.geometry.polyhedron.plot import ProjectionFuncStereographic
            sage: proj = ProjectionFuncStereographic([1.0,1.0])
            sage: proj.__init__([1.0,1.0])
            sage: proj.house
            [ 0.7071067811... -0.7071067811...]
            [ 0.7071067811...  0.7071067811...]
            sage: TestSuite(proj).run(skip='_test_pickling')
        """
        from sage.rings.real_double import RDF

        self.projection_point = vector(projection_point)
        self.dim = self.projection_point.degree()

        pproj = vector(RDF, self.projection_point)
        self.psize = norm(pproj)
        if (self.psize).is_zero():
            raise ValueError("projection direction must be a nonzero vector.")
        v = vector(RDF, [0.0] * (self.dim - 1) + [-self.psize]) - pproj
        polediff = matrix(RDF, v).transpose()
        denom = RDF((polediff.transpose() * polediff)[0][0])
        if denom.is_zero():
            self.house = identity_matrix(RDF, self.dim)
        else:
            house = identity_matrix(RDF, self.dim) \
            - 2*polediff*polediff.transpose()/denom   # Householder reflector
            # Make it preserve orientation (chirality):
            self.house = diagonal_matrix(RDF, [1] * (self.dim - 1) + [-1]) * house

    def __call__(self, x):
        """
        Action of the stereographic projection.

        INPUT:

        - ``x`` -- a vector or anything convertible to a vector

        OUTPUT:

        First reflects ``x`` with a Householder reflection which takes
        the projection point to ``(0,...,0,self.psize)`` where
        ``psize`` is the length of the projection point, and then
        dilates by ``1/(zdiff)`` where ``zdiff`` is the difference
        between the last coordinate of ``x`` and ``psize``.

        EXAMPLES::

            sage: from sage.geometry.polyhedron.plot import ProjectionFuncStereographic
            sage: proj = ProjectionFuncStereographic([1.0,1.0])
            sage: proj.__call__(vector([1,2]))
            (1.0000000000000002)
            sage: proj = ProjectionFuncStereographic([2.0,1.0])
            sage: proj.__call__(vector([1,2]))  # abs tol 1e-14
            (-2.999999999999997)
            sage: proj = ProjectionFuncStereographic([0,0,2])
            sage: proj.__call__(vector([0,0,1]))
            (0.0, 0.0)
            sage: proj.__call__(vector([1,0,0]))
            (0.5, 0.0)
        """
        from sage.rings.real_double import RDF

        img = self.house * x
        denom = self.psize - img[self.dim - 1]
        if denom.is_zero():
            raise ValueError('Point cannot coincide with '
                             'coordinate singularity at ' + repr(x))
        return vector(RDF, [img[i] / denom for i in range(self.dim - 1)])


class ProjectionFuncSchlegel:
    """
    The Schlegel projection from the given input point.

    EXAMPLES::

        sage: from sage.geometry.polyhedron.plot import ProjectionFuncSchlegel
        sage: fcube = polytopes.hypercube(4)
        sage: facet = fcube.facets()[0]
        sage: proj = ProjectionFuncSchlegel(facet,[0,-1.5,0,0])
        sage: proj([0,0,0,0])[0]
        1.0
    """
    def __init__(self, facet, projection_point):
        """
        Initialize the projection.

        EXAMPLES::

            sage: from sage.geometry.polyhedron.plot import ProjectionFuncSchlegel
            sage: fcube = polytopes.hypercube(4)
            sage: facet = fcube.facets()[0]
            sage: proj = ProjectionFuncSchlegel(facet,[0,-1.5,0,0])
            sage: proj.facet
            A 3-dimensional face of a Polyhedron in ZZ^4 defined as the convex hull of 8 vertices
            sage: proj.A
            [1.0 0.0 0.0]
            [0.0 0.0 0.0]
            [0.0 0.0 1.0]
            [0.0 1.0 0.0]
            sage: proj.b
            (1.0, 1.0, 1.0)
            sage: proj.projection_point
            (0.0, -1.5, 0.0, 0.0)
            sage: proj([-1,1,1,1])
            (0.8, 1.2, 1.2)
            sage: proj([1,1,1,1])
            (1.2, 1.2, 1.2)
            sage: proj([1,-1,1,1])
            (2.0, 2.0, 2.0)
            sage: proj([1,-1,-1,1])
            (2.0, 2.0, 0.0)
            sage: TestSuite(proj).run(skip='_test_pickling')
        """
        from sage.rings.real_double import RDF

        self.facet = facet
        ineq = [h for h in facet.ambient_Hrepresentation() if h.is_inequality()][0]
        self.full_A = ineq.A()
        self.full_b = ineq.b()
        A, b = self.facet.as_polyhedron().affine_hull_projection(as_affine_map=True, orthonormal=True, extend=True)
        self.A = A.change_ring(RDF).matrix()
        self.b = b.change_ring(RDF)
        self.projection_point = vector(RDF, projection_point)

    def __call__(self, x):
        """
        Apply the projection to a vector.

        - ``x`` -- a vector or anything convertible to a vector

        EXAMPLES::

            sage: from sage.geometry.polyhedron.plot import ProjectionFuncSchlegel
            sage: fcube = polytopes.hypercube(4)
            sage: facet = fcube.facets()[0]
            sage: proj = ProjectionFuncSchlegel(facet,[0,-1.5,0,0])
            sage: proj.__call__([0,0,0,0])
            (1.0, 1.0, 1.0)
        """
        # The intersection of the segment with the facet
        # See Ziegler's "Lectures on Polytopes" p.133
        vx = vector(x)
        z = (self.full_b)
        a = -(self.full_A)
        y = self.projection_point
        preimage = y + ((z-a*y)/(a*vx-a*y))*(vx - y)
        # The transformation matrix acts on the right:
        return preimage*self.A + self.b

#########################################################################


class Projection(SageObject):
    """
    The projection of a :class:`Polyhedron`.

    This class keeps track of the necessary data to plot the input
    polyhedron.
    """

    def __init__(self, polyhedron, proj=projection_func_identity):
        """
        Initialize the projection of a Polyhedron() object.

        INPUT:

        - ``polyhedron`` -- a ``Polyhedron()`` object

        - ``proj`` -- a projection function for the points

        .. NOTE::

            Once initialized, the polyhedral data is fixed. However, the
            projection can be changed later on.

        EXAMPLES::

            sage: # needs sage.groups
            sage: p = polytopes.icosahedron(exact=False)
            sage: from sage.geometry.polyhedron.plot import Projection
            sage: Projection(p)
            The projection of a polyhedron into 3 dimensions
            sage: def pr_12(x): return [x[1],x[2]]
            sage: Projection(p, pr_12)
            The projection of a polyhedron into 2 dimensions
            sage: Projection(p,  lambda x: [x[1],x[2]] )   # another way of doing the same projection
            The projection of a polyhedron into 2 dimensions
            sage: _.plot()   # plot of the projected icosahedron in 2d                  # needs sage.plot
            Graphics object consisting of 51 graphics primitives
            sage: proj = Projection(p)
            sage: proj.stereographic([1,2,3])
            The projection of a polyhedron into 2 dimensions
            sage: proj.plot()                                                           # needs sage.plot
            Graphics object consisting of 51 graphics primitives
            sage: TestSuite(proj).run(skip='_test_pickling')
        """
        self.parent_polyhedron = polyhedron
        self.coords = Sequence([])
        self.points = Sequence([])
        self.lines = Sequence([])
        self.arrows = Sequence([])
        self.polygons = Sequence([])
        self.polyhedron_ambient_dim = polyhedron.ambient_dim()
        self.polyhedron_dim = polyhedron.dim()

        if polyhedron.ambient_dim() == 2:
            self._init_from_2d(polyhedron)
        elif polyhedron.ambient_dim() == 3:
            self._init_from_3d(polyhedron)
        else:
            self._init_points(polyhedron)
            self._init_lines_arrows(polyhedron)

        self.coords.set_immutable()
        self.points.set_immutable()
        self.lines.set_immutable()
        self.arrows.set_immutable()
        self.polygons.set_immutable()

        self(proj)

    def _repr_(self):
        """
        Return a string describing the projection.

        EXAMPLES::

            sage: p = polytopes.hypercube(3)
            sage: from sage.geometry.polyhedron.plot import Projection
            sage: proj = Projection(p)
            sage: print(proj._repr_())
            The projection of a polyhedron into 3 dimensions
        """
        s = 'The projection of a polyhedron into '
        s += repr(self.dimension) + ' dimensions'
        return s

    def __call__(self, proj=projection_func_identity):
        """
        Apply a projection.

        EXAMPLES::

            sage: # needs sage.groups
            sage: p = polytopes.icosahedron(exact=False)
            sage: from sage.geometry.polyhedron.plot import Projection
            sage: pproj = Projection(p)
            sage: from sage.geometry.polyhedron.plot import ProjectionFuncStereographic
            sage: pproj_stereo = pproj.__call__(proj=ProjectionFuncStereographic([1, 2, 3]))
            sage: sorted(pproj_stereo.polygons)
            [[2, 0, 9],
             [3, 1, 10],
             [4, 0, 8],
             ...
             [11, 1, 3],
             [11, 3, 7],
             [11, 7, 9]]
        """
        self.transformed_coords = Sequence([proj(p) for p in self.coords])
        self._init_dimension()
        return self

    def identity(self):
        """
        Return the identity projection of the polyhedron.

        EXAMPLES::

            sage: # needs sage.groups
            sage: p = polytopes.icosahedron(exact=False)
            sage: from sage.geometry.polyhedron.plot import Projection
            sage: pproj = Projection(p)
            sage: ppid = pproj.identity()
            sage: ppid.dimension
            3
        """
        return self(projection_func_identity)

    def stereographic(self, projection_point=None):
        r"""
        Return the stereographic projection.

        INPUT:

        - ``projection_point`` -- the projection point. This must be
          distinct from the polyhedron's vertices. Default is `(1,0,\dots,0)`.

        EXAMPLES::

            sage: from sage.geometry.polyhedron.plot import Projection
            sage: proj = Projection(polytopes.buckyball()); proj  # long time
            The projection of a polyhedron into 3 dimensions
            sage: proj.stereographic([5,2,3]).plot()    # long time                     # needs sage.plot
            Graphics object consisting of 123 graphics primitives
            sage: Projection(polytopes.twenty_four_cell()).stereographic([2,0,0,0])
            The projection of a polyhedron into 3 dimensions
        """
        if projection_point is None:
            projection_point = [1] + [0]*(self.polyhedron_ambient_dim-1)
        return self(ProjectionFuncStereographic(projection_point))

    def schlegel(self, facet=None, position=None):
        r"""
        Return the Schlegel projection.

        * The facet is orthonormally transformed into its affine hull.

        * The position specifies a point coming out of the barycenter of the
          facet from which the other vertices will be projected into the facet.

        INPUT:

        - ``facet`` -- a PolyhedronFace; the facet into which the Schlegel
          diagram is created. The default is the first facet.

        - ``position`` -- a positive number. Determines a relative distance
          from the barycenter of ``facet``. A value close to 0 will place the
          projection point close to the facet and a large value further away.
          If the given value is too large, an error is returned.
          If no position is given, it takes the midpoint of the possible
          point of views along a line spanned by the barycenter of the facet
          and a valid point outside the facet.

        EXAMPLES::

            sage: cube4 = polytopes.hypercube(4)
            sage: from sage.geometry.polyhedron.plot import Projection
            sage: Projection(cube4).schlegel()
            The projection of a polyhedron into 3 dimensions
            sage: _.plot()                                                              # needs sage.plot
            Graphics3d Object

        The 4-cube with a truncated vertex seen into the resulting tetrahedron
        facet::

            sage: tcube4 = cube4.face_truncation(cube4.faces(0)[0])
            sage: tcube4.facets()[4]
            A 3-dimensional face of a Polyhedron in QQ^4 defined as the convex hull of 4 vertices
            sage: into_tetra = Projection(tcube4).schlegel(tcube4.facets()[4])          # needs sage.symbolic
            sage: into_tetra.plot()                                                     # needs sage.plot sage.symbolic
            Graphics3d Object

        Taking a larger value for the position changes the image::

            sage: into_tetra_far = Projection(tcube4).schlegel(tcube4.facets()[4], 4)   # needs sage.symbolic
            sage: into_tetra_far.plot()                                                 # needs sage.plot sage.symbolic
            Graphics3d Object

        A value which is too large or negative give a projection point that
        sees more than one facet resulting in a error::

            sage: Projection(tcube4).schlegel(tcube4.facets()[4], 5)
            Traceback (most recent call last):
            ...
            ValueError: the chosen position is too large
            sage: Projection(tcube4).schlegel(tcube4.facets()[4], -1)
            Traceback (most recent call last):
            ...
            ValueError: 'position' should be a positive number
        """
        from sage.geometry.polyhedron.face import PolyhedronFace
        from sage.rings.integer_ring import ZZ
        if facet is None:
            facet = self.parent_polyhedron.facets()[0]
        elif not isinstance(facet, PolyhedronFace):
            raise TypeError("{} should be a PolyhedronFace of {}".format(facet, self))
        elif facet.dim() != self.parent_polyhedron.dim() - 1:
            raise ValueError("The face should be a facet of the polyhedron")
        if position is not None and position <= 0:
            raise ValueError("'position' should be a positive number")

        barycenter = ZZ.one() * sum([v.vector() for v in facet.vertices()]) / len(facet.vertices())
        locus_polyhedron = facet.stacking_locus()
        repr_point = locus_polyhedron.representative_point()
        if position is None:
            # Figure out a somehow canonical point of view inside the locus
            # polyhedron
            from sage.geometry.polyhedron.constructor import Polyhedron
            the_ray = Polyhedron(vertices=[barycenter],
                                 rays=[repr_point - barycenter],
                                 backend=locus_polyhedron.backend()) & locus_polyhedron
            projection_point = the_ray.representative_point()
        else:
            projection_point = (1-position)*barycenter + position*repr_point
            if not locus_polyhedron.relative_interior_contains(projection_point):
                raise ValueError("the chosen position is too large")

        return self(ProjectionFuncSchlegel(facet, projection_point))

    def coord_index_of(self, v):
        """
        Convert a coordinate vector to its internal index.

        EXAMPLES::

            sage: p = polytopes.hypercube(3)
            sage: proj = p.projection()
            sage: proj.coord_index_of(vector((1,1,1)))
            2
        """
        try:
            return self.coords.index(v)
        except ValueError:
            self.coords.append(v)
            return len(self.coords)-1

    def coord_indices_of(self, v_list):
        """
        Convert list of coordinate vectors to the corresponding list
        of internal indices.

        EXAMPLES::

            sage: p = polytopes.hypercube(3)
            sage: proj = p.projection()
            sage: proj.coord_indices_of([vector((1,1,1)), vector((1,-1,1))])
            [2, 3]
        """
        return [self.coord_index_of(v) for v in v_list]

    def coordinates_of(self, coord_index_list):
        """
        Given a list of indices, return the projected coordinates.

        EXAMPLES::

            sage: p = polytopes.simplex(4, project=True).projection()
            sage: p.coordinates_of([1])
            [[-0.7071067812, 0.4082482905, 0.2886751346, 0.2236067977]]
        """
        return [self.transformed_coords[i] for i in coord_index_list]

    def _init_dimension(self):
        """
        Internal function: Initialize from polyhedron with
        projected coordinates. Must always be called after
        a coordinate projection.

        TESTS::

            sage: p = polytopes.simplex(2, project=True).projection()
            sage: test = p._init_dimension()
            sage: p.plot.__doc__ == p.render_2d.__doc__
            True
        """
        if self.transformed_coords:
            self.dimension = len(self.transformed_coords[0])
        else:
            self.dimension = 0
        if self.dimension == 0:
            self.plot = self.render_0d
        elif self.dimension == 1:
            self.plot = self.render_1d
        elif self.dimension == 2:
            self.plot = self.render_2d
        elif self.dimension == 3:
            self.plot = self.render_3d
        else:
            try:
                del self.plot
            except AttributeError:
                pass

    def _init_from_2d(self, polyhedron):
        """
        Internal function: Initialize from polyhedron in
        2-dimensional space. The polyhedron could be lower
        dimensional.

        TESTS::

            sage: p = Polyhedron(vertices=[[0,0],[0,1],[1,0],[1,1]])
            sage: proj = p.projection()
            sage: [proj.coordinates_of([i]) for i in proj.points]
            [[[0, 0]], [[0, 1]], [[1, 0]], [[1, 1]]]
            sage: proj._init_from_2d
            <bound method Projection._init_from_2d of The projection
            of a polyhedron into 2 dimensions>
        """
        assert polyhedron.ambient_dim() == 2, "Requires polyhedron in 2d"
        self.dimension = 2
        self._init_points(polyhedron)
        self._init_lines_arrows(polyhedron)
        self._init_area_2d(polyhedron)

    def _init_from_3d(self, polyhedron):
        """
        Internal function: Initialize from polyhedron in
        3-dimensional space. The polyhedron could be
        lower dimensional.

        TESTS::

            sage: p = Polyhedron(vertices=[[0,0,1],[0,1,2],[1,0,3],[1,1,5]])
            sage: proj = p.projection()
            sage: [proj.coordinates_of([i]) for i in proj.points]
            [[[0, 0, 1]], [[0, 1, 2]], [[1, 0, 3]], [[1, 1, 5]]]
            sage: proj._init_from_3d
            <bound method Projection._init_from_3d of The projection
            of a polyhedron into 3 dimensions>
        """
        assert polyhedron.ambient_dim() == 3, "Requires polyhedron in 3d"
        self.dimension = 3
        self._init_points(polyhedron)
        self._init_lines_arrows(polyhedron)
        self._init_solid_3d(polyhedron)

    def _init_points(self, polyhedron):
        """
        Internal function: Initialize points (works in arbitrary
        dimensions).

        TESTS::

            sage: p = polytopes.hypercube(2)
            sage: pp = p.projection()
            sage: del pp.points
            sage: pp.points = Sequence([])
            sage: pp._init_points(p)
            sage: pp.points
            [0, 1, 2, 3]
        """
        for v in polyhedron.vertex_generator():
            self.points.append(self.coord_index_of(v.vector()))

    def _init_lines_arrows(self, polyhedron):
        """
        Internal function: Initialize compact and non-compact edges
        (works in arbitrary dimensions).

        TESTS::

            sage: p = Polyhedron(ieqs=[[1, 0, 0, 1], [1, 1, 0, 0]])
            sage: pp = p.projection()
            sage: pp.arrows
            [[0, 1], [0, 2], [0, 3], [0, 4]]
            sage: del pp.arrows
            sage: pp.arrows = Sequence([])
            sage: pp._init_lines_arrows(p)
            sage: pp.arrows
            [[0, 1], [0, 2], [0, 3], [0, 4]]

        We check that :issue:`31802` is fixed::

            sage: x = Polyhedron(lines=[(1, 0, 0), (0, 1, 0)], rays=[(0, 0, 1)])
            sage: y = x.projection()
            sage: del y.arrows
            sage: y.arrows = Sequence([])
            sage: y._init_lines_arrows(x)
            sage: y.arrows
            [[0, 1], [0, 2], [0, 3], [0, 4], [0, 5]]
        """
        obj = polyhedron.Vrepresentation()
        adj_matrix = polyhedron.vertex_adjacency_matrix()
        for i in range(len(obj)):
            if not obj[i].is_vertex():
                if any(adj_matrix[i, j] != 0
                       for j in range(len(obj))):
                    continue
                # obj[i] is ray or line
                v = polyhedron.vertices()[0].vector()
                r = obj[i].vector()
                self.arrows.append( [ self.coord_index_of(v),
                                      self.coord_index_of(v + r) ] )
                if obj[i].is_line():
                    self.arrows.append( [ self.coord_index_of(v),
                                          self.coord_index_of(v - r) ] )
            for j in range(len(obj)):
                if adj_matrix[i, j] == 0:
                    continue
                if i < j and obj[j].is_vertex():
                    l = [obj[i].vector(), obj[j].vector()]
                    self.lines.append( [ self.coord_index_of(l[0]),
                                         self.coord_index_of(l[1]) ] )
                if obj[j].is_ray():
                    l = [obj[i].vector(), obj[i].vector() + obj[j].vector()]
                    self.arrows.append( [ self.coord_index_of(l[0]),
                                          self.coord_index_of(l[1]) ] )
                if obj[j].is_line():
                    l1 = [obj[i].vector(), obj[i].vector() + obj[j].vector()]
                    l2 = [obj[i].vector(), obj[i].vector() - obj[j].vector()]
                    self.arrows.append( [ self.coord_index_of(l1[0]),
                                          self.coord_index_of(l1[1]) ] )
                    self.arrows.append( [ self.coord_index_of(l2[0]),
                                          self.coord_index_of(l2[1]) ] )

    def _init_area_2d(self, polyhedron):
        """
        Internal function: Initialize polygon area for 2d polyhedron.

        TESTS::

            sage: p = polytopes.cyclic_polytope(2,4)
            sage: proj = p.projection()
            sage: proj.polygons = Sequence([])
            sage: proj._init_area_2d(p)
            sage: proj.polygons
            [[3, 0, 1, 2]]
        """
        assert polyhedron.ambient_dim() == 2, "Requires polyhedron in 2d"
        vertices = list(polyhedron.Vrep_generator())
        vertices = cyclic_sort_vertices_2d(vertices)
        coords = []

        def adjacent_vertices(i):
            n = len(vertices)
            if vertices[(i-1) % n].is_vertex():
                yield vertices[(i-1) % n]
            if vertices[(i+1) % n].is_vertex():
                yield vertices[(i+1) % n]

        for i in range(len(vertices)):
            v = vertices[i]
            if v.is_vertex():
                coords.append(v())
            if v.is_ray():
                for a in adjacent_vertices(i):
                    coords.append(a() + v())

        if polyhedron.n_lines() == 0:
            self.polygons.append(self.coord_indices_of(coords))
            return

        polygons = []

        if polyhedron.n_lines() == 1:
            a_line = next(polyhedron.line_generator())
            for shift in [a_line(), -a_line()]:
                for i in range(len(coords)):
                    polygons.append([coords[i - 1], coords[i],
                                     coords[i] + shift, coords[i - 1] + shift])

        if polyhedron.n_lines() == 2:
            line1, line2 = polyhedron.lines()
            assert len(coords) == 1, "Can have only a single vertex!"
            v = coords[0]
            l1 = line1()
            l2 = line2()
            polygons = [[v-l1-l2, v+l1-l2, v+l1+l2, v-l1+l2]]

        polygons = [self.coord_indices_of(p) for p in polygons]
        self.polygons.extend(polygons)

    def _init_solid_3d(self, polyhedron):
        """
        Internal function: Initialize facet polygons for 3d polyhedron.

        TESTS::

            sage: p = polytopes.cyclic_polytope(3,4)
            sage: proj = p.projection()
            sage: proj.polygons = Sequence([])
            sage: proj._init_solid_3d(p)
            sage: proj.polygons
            [[1, 0, 2], [3, 0, 1], [2, 0, 3], [3, 1, 2]]

            sage: x = Polyhedron(rays=[(-1, 0, 0), (1, 0, 0), (0, 1, 0), (0, 0, 1)])
            sage: y = x.projection()
            sage: y.polygons
            [[5, 2, 1, 6], [2, 7, 8, 1]]

            sage: cylinder = Polyhedron(vertices=[(0, 0, 0), (1, 0, 0), (0, 1, 0)], lines=[(0, 0, 1)])
            sage: len(cylinder.projection().polygons)
            3
        """
        assert polyhedron.ambient_dim() == 3, "Requires polyhedron in 3d"

        if polyhedron.dim() <= 1:  # empty or 0d or 1d polyhedron => no polygon
            return None

        def defining_equation():  # corresponding to a polygon
            if polyhedron.dim() < 3:
                yield next(polyhedron.equation_generator())
            else:
                yield from polyhedron.inequality_generator()

        faces = []
        face_inequalities = []
        for facet_equation in defining_equation():
            vertices = list(facet_equation.incident())
            face_inequalities.append(facet_equation)
            vertices = cyclic_sort_vertices_2d(vertices)
            if len(vertices) >= 3:
                v0, v1, v2 = (vector(v) for v in vertices[:3])
                normal = (v2 - v0).cross_product(v1 - v0)
                if normal.dot_product(facet_equation.A()) < 0:
                    vertices.reverse()
            coords = []

            def adjacent_vertices(i):
                n = len(vertices)
                if vertices[(i-1) % n].is_vertex():
                    yield vertices[(i-1) % n]
                if vertices[(i+1) % n].is_vertex():
                    yield vertices[(i+1) % n]

            for i in range(len(vertices)):
                v = vertices[i]
                if v.is_vertex():
                    coords.append(v())
                if v.is_ray():
                    for a in adjacent_vertices(i):
                        coords.append(a() + v())

            faces.append(coords)
        self.face_inequalities = face_inequalities

        if polyhedron.n_lines() == 0:
            assert faces, "no vertices?"
            self.polygons.extend( [self.coord_indices_of(f) for f in faces] )
            return

        # now some special cases if there are lines (dim < ambient_dim)
        polygons = []

        if polyhedron.n_lines() == 1:
            assert faces, "no vertices?"
            a_line = next(polyhedron.line_generator())
            shift = a_line()
            for coords in faces:
                assert len(coords) == 2, "There must be two points."
                polygons.append([coords[0] - shift, coords[1] - shift,
                                 coords[1] + shift, coords[0] + shift])

        if polyhedron.n_lines() == 2:
            line1, line2 = polyhedron.line_generator()
            l1 = line1()
            l2 = line2()
            for v in polyhedron.vertex_generator():
                polygons.append([v()-l1-l2, v()+l1-l2, v()+l1+l2, v()-l1+l2])

        self.polygons.extend([self.coord_indices_of(p) for p in polygons])

    def render_points_1d(self, **kwds):
        """
        Return the points of a polyhedron in 1d.

        INPUT:

        - ``**kwds`` -- options passed through to
          :func:`~sage.plot.point.point2d`

        OUTPUT: a 2-d graphics object

        EXAMPLES::

            sage: cube1 = polytopes.hypercube(1)
            sage: proj = cube1.projection()
            sage: points = proj.render_points_1d()                                      # needs sage.plot
            sage: points._objects                                                       # needs sage.plot
            [Point set defined by 2 point(s)]
        """
        return point2d([c + [0] for c in self.coordinates_of(self.points)], **kwds)

    def render_line_1d(self, **kwds):
        """
        Return the line of a polyhedron in 1d.

        INPUT:

        - ``**kwds`` -- options passed through to
          :func:`~sage.plot.line.line2d`

        OUTPUT: a 2-d graphics object

        EXAMPLES::

            sage: outline = polytopes.hypercube(1).projection().render_line_1d()        # needs sage.plot
            sage: outline._objects[0]                                                   # needs sage.plot
            Line defined by 2 points
        """
        if len(self.lines) == 0:
            return Graphics()
        elif len(self.lines) == 1:
            line = self.coordinates_of(self.lines[0])
            return line2d([line[0] + [0], line[1] + [0]], **kwds)
        else:
            assert False   # unreachable

    def render_points_2d(self, **kwds):
        """
        Return the points of a polyhedron in 2d.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: hex = polytopes.regular_polygon(6)
            sage: proj = hex.projection()
            sage: hex_points = proj.render_points_2d()                                  # needs sage.plot
            sage: hex_points._objects                                                   # needs sage.plot
            [Point set defined by 6 point(s)]
        """
        return point2d(self.coordinates_of(self.points), **kwds)

    def render_outline_2d(self, **kwds):
        """
        Return the outline (edges) of a polyhedron in 2d.

        EXAMPLES::

            sage: penta = polytopes.regular_polygon(5)                                  # needs sage.rings.number_field
            sage: outline = penta.projection().render_outline_2d()                      # needs sage.plot sage.rings.number_field
            sage: outline._objects[0]                                                   # needs sage.plot sage.rings.number_field
            Line defined by 2 points
        """
        wireframe = []
        for l in self.lines:
            l_coords = self.coordinates_of(l)
            wireframe.append(line2d(l_coords, **kwds))
        for a in self.arrows:
            a_coords = self.coordinates_of(a)
            wireframe.append(arrow(a_coords[0], a_coords[1], **kwds))
        return sum(wireframe)

    def render_fill_2d(self, **kwds):
        """
        Return the filled interior (a polygon) of a polyhedron in 2d.

        EXAMPLES::

            sage: cps = [i^3 for i in srange(-2, 2, 1/5)]
            sage: p = Polyhedron(vertices=[[(t^2-1)/(t^2+1), 2*t/(t^2+1)] for t in cps])
            sage: proj = p.projection()
            sage: filled_poly = proj.render_fill_2d()                                   # needs sage.plot
            sage: filled_poly.axes_width()                                              # needs sage.plot
            0.8
        """
        poly = [polygon2d(self.coordinates_of(p), **kwds)
                for p in self.polygons]
        return sum(poly)

    def render_vertices_3d(self, **kwds):
        """
        Return the 3d rendering of the vertices.

        EXAMPLES::

            sage: p = polytopes.cross_polytope(3)
            sage: proj = p.projection()
            sage: verts = proj.render_vertices_3d()                                     # needs sage.plot
            sage: verts.bounding_box()                                                  # needs sage.plot
            ((-1.0, -1.0, -1.0), (1.0, 1.0, 1.0))
        """
        return point3d(self.coordinates_of(self.points), **kwds)

    def render_wireframe_3d(self, **kwds):
        r"""
        Return the 3d wireframe rendering.

        EXAMPLES::

            sage: cube = polytopes.hypercube(3)
            sage: cube_proj = cube.projection()
            sage: wire = cube_proj.render_wireframe_3d()                                # needs sage.plot
            sage: print(wire.tachyon().split('\n')[77])  # for testing                  # needs sage.plot
            FCylinder base 1.0 1.0 -1.0 apex -1.0 1.0 -1.0 rad 0.005 texture...
        """
        wireframe = []
        for l in self.lines:
            l_coords = self.coordinates_of(l)
            wireframe.append(line3d(l_coords, **kwds))
        for a in self.arrows:
            a_coords = self.coordinates_of(a)
            wireframe.append(arrow3d(a_coords[0], a_coords[1], **kwds))
        return sum(wireframe)

    def render_solid_3d(self, **kwds):
        """
        Return solid 3d rendering of a 3d polytope.

        EXAMPLES::

            sage: p = polytopes.hypercube(3).projection()
            sage: p_solid = p.render_solid_3d(opacity=.7)                               # needs sage.plot
            sage: type(p_solid)                                                         # needs sage.plot
            <class 'sage.plot.plot3d.index_face_set.IndexFaceSet'>
        """
        polys = self.polygons
        n = len(polys)
        N = max([-1] + [i for p in polys for i in p]) + 1
        coords = self.coordinates_of(range(N))
        col = kwds.pop('color', (0, 0, 1))
        if col == 'rainbow':
            t_list = [Texture(rainbow(n, 'rgbtuple')[i]) for i in range(n)]
            return polygons3d(polys, coords, texture_list=t_list, **kwds)
        else:
            return polygons3d(polys, coords, color=col, **kwds)

    def render_0d(self, point_opts=None, line_opts=None, polygon_opts=None):
        """
        Return 0d rendering of the projection of a polyhedron into
        2-dimensional ambient space.

        INPUT:

        See
        :meth:`~sage.geometry.polyhedron.base.Polyhedron_base.plot`.

        OUTPUT: a 2-d graphics object

        EXAMPLES::

            sage: print(Polyhedron([]).projection().render_0d().description())          # needs sage.plot
            <BLANKLINE>
            sage: P = Polyhedron(ieqs=[(1,)])
            sage: print(P.projection().render_0d().description())                       # needs sage.plot
            Point set defined by 1 point(s):    [(0.0, 0.0)]
        """
        if point_opts is None:
            point_opts = {}
        if isinstance(point_opts, dict):
            point_opts.setdefault('zorder', 2)
            point_opts.setdefault('pointsize', 10)
        if self.points:
            return point2d([0, 0], **point_opts)
        else:
            return Graphics()

    def render_1d(self, point_opts=None, line_opts=None, polygon_opts=None):
        """
        Return 1d rendering of the projection of a polyhedron into
        2-dimensional ambient space.

        INPUT:

        See
        :meth:`~sage.geometry.polyhedron.base.Polyhedron_base.plot`.

        OUTPUT: a 2-d graphics object

        EXAMPLES::

            sage: Polyhedron([(0,), (1,)]).projection().render_1d()                     # needs sage.plot
            Graphics object consisting of 2 graphics primitives
        """
        plt = Graphics()
        if point_opts is None:
            point_opts = {}
        if line_opts is None:
            line_opts = {}
        if isinstance(point_opts, dict):
            point_opts.setdefault('zorder', 2)
            point_opts.setdefault('pointsize', 10)
            plt += self.render_points_1d(**point_opts)
        if isinstance(line_opts, dict):
            line_opts.setdefault('zorder', 1)
            plt += self.render_line_1d(**line_opts)
        return plt

    def render_2d(self, point_opts=None, line_opts=None, polygon_opts=None):
        """
        Return 2d rendering of the projection of a polyhedron into
        2-dimensional ambient space.

        EXAMPLES::

            sage: p1 = Polyhedron(vertices=[[1,1]], rays=[[1,1]])
            sage: q1 = p1.projection()
            sage: p2 = Polyhedron(vertices=[[1,0], [0,1], [0,0]])
            sage: q2 = p2.projection()
            sage: p3 = Polyhedron(vertices=[[1,2]])
            sage: q3 = p3.projection()
            sage: p4 = Polyhedron(vertices=[[2,0]], rays=[[1,-1]], lines=[[1,1]])
            sage: q4 = p4.projection()
            sage: q1.plot() + q2.plot() + q3.plot() + q4.plot()                         # needs sage.plot
            Graphics object consisting of 18 graphics primitives
        """
        plt = Graphics()
        if point_opts is None:
            point_opts = {}
        if line_opts is None:
            line_opts = {}
        if polygon_opts is None:
            polygon_opts = {}
        if isinstance(point_opts, dict):
            point_opts.setdefault('zorder', 2)
            point_opts.setdefault('pointsize', 10)
            plt += self.render_points_2d(**point_opts)
        if isinstance(line_opts, dict):
            line_opts.setdefault('zorder', 1)
            plt += self.render_outline_2d(**line_opts)
        if isinstance(polygon_opts, dict):
            polygon_opts.setdefault('zorder', 0)
            plt += self.render_fill_2d(**polygon_opts)
        return plt

    def render_3d(self, point_opts=None, line_opts=None, polygon_opts=None):
        """
        Return 3d rendering of a polyhedron projected into
        3-dimensional ambient space.

        EXAMPLES::

            sage: p1 = Polyhedron(vertices=[[1,1,1]], rays=[[1,1,1]])
            sage: p2 = Polyhedron(vertices=[[2,0,0], [0,2,0], [0,0,2]])
            sage: p3 = Polyhedron(vertices=[[1,0,0], [0,1,0], [0,0,1]],
            ....:                 rays=[[-1,-1,-1]])
            sage: (p1.projection().plot() + p2.projection().plot()                      # needs sage.plot
            ....:   + p3.projection().plot())
            Graphics3d Object

        It correctly handles various degenerate cases::

            sage: # needs sage.plot
            sage: Polyhedron(lines=[[1,0,0], [0,1,0], [0,0,1]]).plot()  # whole space
            Graphics3d Object
            sage: Polyhedron(vertices=[[1,1,1]], rays=[[1,0,0]],
            ....:            lines=[[0,1,0], [0,0,1]]).plot()           # half space
            Graphics3d Object
            sage: Polyhedron(lines=[[0,1,0], [0,0,1]],
            ....:            vertices=[[1,1,1]]).plot()      # R^2 in R^3
            Graphics3d Object
            sage: Polyhedron(rays=[[0,1,0], [0,0,1]],        # quadrant wedge in R^2
            ....:            lines=[[1,0,0]]).plot()
            Graphics3d Object
            sage: Polyhedron(rays=[[0,1,0]],                 # upper half plane in R^3
            ....:            lines=[[1,0,0]]).plot()
            Graphics3d Object
            sage: Polyhedron(lines=[[1,0,0]]).plot()         # R^1 in R^2
            Graphics3d Object
            sage: Polyhedron(rays=[[0,1,0]]).plot()          # Half-line in R^3
            Graphics3d Object
            sage: Polyhedron(vertices=[[1,1,1]]).plot()      # point in R^3
            Graphics3d Object

        The origin is not included, if it is not in the polyhedron (:issue:`23555`)::

            sage: Q = Polyhedron([[100],[101]])
            sage: P = Q*Q*Q; P
            A 3-dimensional polyhedron in ZZ^3 defined as the convex hull of 8 vertices
            sage: p = P.plot()                                                          # needs sage.plot
            sage: p.bounding_box()                                                      # needs sage.plot
            ((100.0, 100.0, 100.0), (101.0, 101.0, 101.0))

        Plot 3d polytope with rainbow colors::

            sage: polytopes.hypercube(3).plot(polygon='rainbow', alpha=0.4)             # needs sage.plot
            Graphics3d Object
        """
        pplt = None
        lplt = None
        pgplt = None
        if point_opts is None:
            point_opts = {}
        if line_opts is None:
            line_opts = {}
        if polygon_opts is None:
            polygon_opts = {}
        if isinstance(point_opts, dict):
            point_opts.setdefault('size', 10)
            pplt = self.render_vertices_3d(**point_opts)
        if isinstance(line_opts, dict):
            line_opts.setdefault('width', 1) # controls the width of arrow3d for a ray
            line_opts.setdefault('thickness', 1) # controls the thickness of line3d
            lplt = self.render_wireframe_3d(**line_opts)
        if isinstance(polygon_opts, dict):
            if 'threejs_flat_shading' not in polygon_opts:
                polygon_opts['threejs_flat_shading'] = True
            pgplt = self.render_solid_3d(**polygon_opts)
        # zorder is not available
        return sum(_ for _ in [pplt, lplt, pgplt] if _ is not None)

    def tikz(self, view=[0, 0, 1], angle=0, scale=1,
             edge_color='blue!95!black', facet_color='blue!95!black',
             opacity=0.8, vertex_color='green', axis=False,
             output_type=None):
        r"""
        Return a tikz picture of ``self`` as a string or as a
        :class:`~sage.misc.latex_standalone.TikzPicture`
        according to a projection ``view`` and an angle ``angle``
        obtained via the threejs viewer.

        INPUT:

        - ``view`` -- list (default: [0,0,1]) representing the rotation axis (see note below)
        - ``angle`` -- integer (default: 0); angle of rotation in degree from 0 to 360 (see note
          below)
        - ``scale`` -- integer (default: 1); the scaling of the tikz picture
        - ``edge_color`` -- string (default: ``'blue!95!black'``); representing colors which tikz
          recognizes
        - ``facet_color`` -- string (default: ``'blue!95!black'``); representing colors which tikz
          recognizes
        - ``vertex_color`` -- string (default: ``'green'``); representing colors which tikz
          recognizes
        - ``opacity`` -- real number (default: 0.8) between 0 and 1 giving the opacity of
          the front facets
        - ``axis`` -- boolean (default: ``False``); draw the axes at the origin or not
        - ``output_type`` -- string (default: ``None``); valid values
          are ``None`` (deprecated), ``'LatexExpr'`` and ``'TikzPicture'``,
          whether to return a :class:`LatexExpr` object (which inherits from Python
          :class:`str`) or a :class:`TikzPicture` object from module
          :mod:`sage.misc.latex_standalone`

        OUTPUT: :class:`LatexExpr` object or :class:`TikzPicture` object

        .. NOTE::

            The inputs ``view`` and ``angle`` can be obtained by visualizing it
            using ``.show(aspect_ratio=1)``. This will open an interactive view
            in your default browser, where you can rotate the polytope. Once
            the desired view angle is found, click on the information icon in
            the lower right-hand corner and select *Get Viewpoint*. This will
            copy a string of the form '[x,y,z],angle' to your local clipboard.
            Go back to Sage and type ``Img = P.projection().tikz([x,y,z],angle)``.

            The inputs ``view`` and ``angle`` can also be obtained from the
            viewer Jmol::

                1) Right click on the image
                2) Select ``Console``
                3) Select the tab ``State``
                4) Scroll to the line ``moveto``

            It reads something like::

                moveto 0.0 {x y z angle} Scale

            The ``view`` is then [x,y,z] and ``angle`` is angle.
            The following number is the scale.

            Jmol performs a rotation of ``angle`` degrees along the
            vector [x,y,z] and show the result from the z-axis.

        EXAMPLES::

            sage: # needs sage.plot sage.rings.number_field
            sage: P1 = polytopes.small_rhombicuboctahedron()
            sage: Image1 = P1.projection().tikz([1,3,5], 175, scale=4,
            ....:                               output_type='TikzPicture')
            sage: type(Image1)
            <class 'sage.misc.latex_standalone.TikzPicture'>
            sage: Image1
            \documentclass[tikz]{standalone}
            \begin{document}
            \begin{tikzpicture}%
                    [x={(-0.939161cm, 0.244762cm)},
                    y={(0.097442cm, -0.482887cm)},
                    z={(0.329367cm, 0.840780cm)},
                    scale=4.000000,
            ...
            Use print to see the full content.
            ...
            \node[vertex] at (-2.41421, 1.00000, -1.00000)     {};
            \node[vertex] at (-2.41421, -1.00000, 1.00000)     {};
            %%
            %%
            \end{tikzpicture}
            \end{document}
            sage: _ = Image1.tex('polytope-tikz1.tex')          # not tested
            sage: _ = Image1.png('polytope-tikz1.png')          # not tested
            sage: _ = Image1.pdf('polytope-tikz1.pdf')          # not tested
            sage: _ = Image1.svg('polytope-tikz1.svg')          # not tested

        A second example::

            sage: P2 = Polyhedron(vertices=[[1, 1], [1, 2], [2, 1]])
            sage: Image2 = P2.projection().tikz(scale=3, edge_color='blue!95!black',
            ....:                               facet_color='orange!95!black', opacity=0.4,
            ....:                               vertex_color='yellow', axis=True,
            ....:                               output_type='TikzPicture')
            sage: Image2
            \documentclass[tikz]{standalone}
            \begin{document}
            \begin{tikzpicture}%
                    [scale=3.000000,
                    back/.style={loosely dotted, thin},
                    edge/.style={color=blue!95!black, thick},
                    facet/.style={fill=orange!95!black,fill opacity=0.400000},
            ...
            Use print to see the full content.
            ...
            \node[vertex] at (1.00000, 2.00000)     {};
            \node[vertex] at (2.00000, 1.00000)     {};
            %%
            %%
            \end{tikzpicture}
            \end{document}

        The second example using a LatexExpr as output type::

            sage: # needs sage.plot
            sage: Image2 = P2.projection().tikz(scale=3, edge_color='blue!95!black',
            ....:                               facet_color='orange!95!black', opacity=0.4,
            ....:                               vertex_color='yellow', axis=True,
            ....:                               output_type='LatexExpr')
            sage: type(Image2)
            <class 'sage.misc.latex.LatexExpr'>
            sage: print('\n'.join(Image2.splitlines()[:4]))
            \begin{tikzpicture}%
                [scale=3.000000,
                back/.style={loosely dotted, thin},
                edge/.style={color=blue!95!black, thick},
            sage: with open('polytope-tikz2.tex', 'w') as f:    # not tested
            ....:     _ = f.write(Image2)

        A third example::

            sage: # needs sage.plot
            sage: P3 = Polyhedron(vertices=[[-1, -1, 2], [-1, 2, -1], [2, -1, -1]]); P3
            A 2-dimensional polyhedron in ZZ^3 defined as the convex hull of 3 vertices
            sage: Image3 = P3.projection().tikz([0.5, -1, -0.1], 55, scale=3,
            ....:                               edge_color='blue!95!black',
            ....:                               facet_color='orange!95!black', opacity=0.7,
            ....:                               vertex_color='yellow', axis=True,
            ....:                               output_type='TikzPicture')
            sage: Image3
            \documentclass[tikz]{standalone}
            \begin{document}
            \begin{tikzpicture}%
                    [x={(0.658184cm, -0.242192cm)},
                    y={(-0.096240cm, 0.912008cm)},
                    z={(-0.746680cm, -0.331036cm)},
                    scale=3.000000,
            ...
            Use print to see the full content.
            ...
            \node[vertex] at (-1.00000, 2.00000, -1.00000)     {};
            \node[vertex] at (2.00000, -1.00000, -1.00000)     {};
            %%
            %%
            \end{tikzpicture}
            \end{document}
            sage: _ = Image3.tex('polytope-tikz3.tex')          # not tested
            sage: _ = Image3.png('polytope-tikz3.png')          # not tested
            sage: _ = Image3.pdf('polytope-tikz3.pdf')          # not tested
            sage: _ = Image3.svg('polytope-tikz3.svg')          # not tested

        A fourth example::

            sage: P = Polyhedron(vertices=[[1,1,0,0], [1,2,0,0],
            ....:                          [2,1,0,0], [0,0,1,0], [0,0,0,1]]); P
            A 4-dimensional polyhedron in ZZ^4 defined as the convex hull of 5 vertices
            sage: P.projection().tikz(output_type='TikzPicture')
            Traceback (most recent call last):
            ...
            NotImplementedError: The polytope has to live in 2 or 3 dimensions.

        .. TODO::

            Make it possible to draw Schlegel diagram for 4-polytopes. ::

                sage: P = Polyhedron(vertices=[[1,1,0,0], [1,2,0,0],
                ....:                          [2,1,0,0], [0,0,1,0], [0,0,0,1]]); P
                A 4-dimensional polyhedron in ZZ^4 defined as the convex hull of 5 vertices
                sage: P.projection().tikz(output_type='TikzPicture')
                Traceback (most recent call last):
                ...
                NotImplementedError: The polytope has to live in 2 or 3 dimensions.

            Make it possible to draw 3-polytopes living in higher dimension.
        """
        if self.polyhedron_ambient_dim > 3 or self.polyhedron_ambient_dim < 2:
            raise NotImplementedError("The polytope has to live in 2 or 3 dimensions.")
        elif self.polyhedron_dim < 2 or self.polyhedron_dim > 3:
            raise NotImplementedError("The polytope has to be 2 or 3-dimensional.")
        elif self.polyhedron_ambient_dim == 2:  # self is a polygon in 2-space
            tikz_string = self._tikz_2d(scale, edge_color, facet_color, opacity,
                                 vertex_color, axis)
        elif self.polyhedron_dim == 2:  # self is a polygon in 3-space
            tikz_string = self._tikz_2d_in_3d(view, angle, scale, edge_color,
                                       facet_color, opacity, vertex_color, axis)
        else:  # self is a 3-polytope in 3-space
            tikz_string = self._tikz_3d_in_3d(view, angle, scale, edge_color,
                                       facet_color, opacity, vertex_color, axis)

        # set default value
        if output_type is None:
            from sage.misc.superseded import deprecation
            msg = ("The default type of the returned object will soon be "
                   "changed from `sage.misc.latex.LatexExpr` to "
                   "`sage.misc.latex_standalone.TikzPicture`.  Please "
                   "update your code to specify the desired output type as "
                   "`.tikz(output_type='LatexExpr')` to keep the old "
                   "behavior or `.tikz(output_type='TikzPicture')` to use "
                   "the future default behavior.")
            deprecation(33002, msg)
            output_type = 'LatexExpr'

        # return
        if output_type == 'LatexExpr':
            return tikz_string
        elif output_type == 'TikzPicture':
            from sage.misc.latex_standalone import TikzPicture
            return TikzPicture(tikz_string, standalone_config=None,
                    usepackage=None, usetikzlibrary=None, macros=None,
                    use_sage_preamble=False)
        else:
            raise ValueError("output_type (='{}') must be 'LatexExpr' or"
                    " 'TikzPicture'".format(output_type))

    def _tikz_2d(self, scale, edge_color, facet_color, opacity, vertex_color, axis):
        r"""
        Return a string ``tikz_pic`` consisting of a tikz picture of
        ``self``, which is assumed to be a polygon on the plane.

        INPUT:

        - ``scale`` -- integer specifying the scaling of the tikz picture
        - ``edge_color`` -- string representing colors which tikz
          recognizes
        - ``facet_color`` -- string representing colors which tikz
          recognizes
        - ``vertex_color`` -- string representing colors which tikz
          recognizes
        - ``opacity`` -- real number between 0 and 1 giving the opacity of
          the front facets
        - ``axis`` -- boolean (default: ``False``); draw the axes at the origin or not

        OUTPUT:

        :class:`LatexExpr` -- containing the TikZ picture.

        EXAMPLES::

            sage: P = Polyhedron(vertices=[[1, 1], [1, 2], [2, 1]])
            sage: Image = P.projection()._tikz_2d(scale=3, edge_color='black',
            ....:                                 facet_color='orange', opacity=0.75,
            ....:                                 vertex_color='yellow', axis=True)
            sage: type(Image)
            <class 'sage.misc.latex.LatexExpr'>
            sage: print('\n'.join(Image.splitlines()[:4]))
            \begin{tikzpicture}%
                [scale=3.000000,
                back/.style={loosely dotted, thin},
                edge/.style={color=black, thick},
            sage: with open('polytope-tikz2.tex', 'w') as f:  # not tested
            ....:     _ = f.write(Image)

        Scientific notation is not used in the output (:issue:`16519`)::

            sage: P = Polyhedron([[2*10^-10,0], [0,1], [1,0]], base_ring=QQ)
            sage: tikz = P.projection().tikz(output_type='TikzPicture')
            sage: 'e-10' in tikz.content()
            False

        .. NOTE::

            The ``facet_color`` is the filing color of the polytope (polygon).
        """

        # Creates the nodes, coordinate and tag for every vertex of the polytope.
        # The tag is used to draw the front facets later on.

        dict_drawing = {}
        edges = ''
        for vert in self.points:
            v = self.coords[vert]
            v_vect = str(['%.5f' % i for i in v]).replace('\'', '')
            v_vect = v_vect.replace('[', '(')
            v_vect = v_vect.replace(']', ')')
            tag = '%s' % v_vect
            node = "\\node[%s] at %s     {};\n" % ('vertex', tag)
            coord = '\\coordinate %s at %s;\n' % (tag, tag)
            dict_drawing[vert] = node, coord, tag

        for index1, index2 in self.lines:
            # v1 = self.coords[index1]
            # v2 = self.coords[index2]
            edges += "\\draw[%s] %s -- %s;\n" % ('edge',
                                                 dict_drawing[index1][2],
                                                 dict_drawing[index2][2])

        # Start to write the output
        tikz_pic = ''
        tikz_pic += '\\begin{tikzpicture}%\n'
        tikz_pic += '\t[scale=%f,\n' % scale
        tikz_pic += '\tback/.style={loosely dotted, thin},\n'
        tikz_pic += '\tedge/.style={color=%s, thick},\n' % edge_color
        tikz_pic += '\tfacet/.style={fill=%s,fill opacity=%f},\n' % (facet_color, opacity)
        tikz_pic += '\tvertex/.style={inner sep=1pt,circle,draw=%s!25!black,' % vertex_color
        tikz_pic += 'fill=%s!75!black,thick}]\n%%\n%%\n' % vertex_color

        # Gives the reproduction information
        from sage.env import SAGE_VERSION
        tikz_pic += "%% This TikZ-picture was produced with Sagemath version {}\n".format(SAGE_VERSION)
        tikz_pic += "%% with the command: ._tikz_2d and parameters:\n"
        tikz_pic += "%% scale = {}\n".format(scale)
        tikz_pic += "%% edge_color = {}\n".format(edge_color)
        tikz_pic += "%% facet_color = {}\n".format(facet_color)
        tikz_pic += "%% opacity = {}\n".format(opacity)
        tikz_pic += "%% vertex_color = {}\n".format(vertex_color)
        tikz_pic += "%% axis = {}\n%%\n".format(axis)

        # Draws the axes if True
        if axis:
            tikz_pic += '%% Drawing the axes\n'
            tikz_pic += '\\draw[color=black,thick,->] (0,0,0) -- (1,0,0) node[anchor=north east]{$x$};\n'
            tikz_pic += '\\draw[color=black,thick,->] (0,0,0) -- (0,1,0) node[anchor=north west]{$y$};\n\n'

        # Create the coordinate of the vertices:
        tikz_pic += '%% Coordinate of the vertices:\n%%\n'
        for v in dict_drawing:
            tikz_pic += dict_drawing[v][1]

        # Draw the interior by going in a cycle
        vertices = list(self.parent_polyhedron.Vrep_generator())
        tikz_pic += '%%\n%%\n%% Drawing the interior\n%%\n'
        cyclic_vert = cyclic_sort_vertices_2d(list(self.parent_polyhedron.Vrep_generator()))
        cyclic_indices = [vertices.index(v) for v in cyclic_vert]
        tikz_pic += '\\fill[facet] '
        for v in cyclic_indices:
            if v in dict_drawing:
                tikz_pic += '%s -- ' % dict_drawing[v][2]
        tikz_pic += 'cycle {};\n'

        # Draw the edges
        tikz_pic += '%%\n%%\n%% Drawing edges\n%%\n'
        tikz_pic += edges

        # Finally, the vertices in front are drawn on top of everything.
        tikz_pic += '%%\n%%\n%% Drawing the vertices\n%%\n'
        for v in dict_drawing:
            tikz_pic += dict_drawing[v][0]
        tikz_pic += '%%\n%%\n\\end{tikzpicture}'

        return LatexExpr(tikz_pic)

    def _tikz_2d_in_3d(self, view, angle, scale, edge_color, facet_color,
                       opacity, vertex_color, axis):
        r"""
        Return a string ``tikz_pic`` consisting of a tikz picture of ``self``
        according to a projection ``view`` and an angle ``angle``
        obtained via Jmol through the current state property. ``self`` is
        assumed to be a polygon in 3-space.

        INPUT:

        - ``view`` -- list (default: [0,0,1]) representing the rotation axis
        - ``angle`` -- integer angle of rotation in degree from 0 to 360
        - ``scale`` -- integer specifying the scaling of the tikz picture
        - ``edge_color`` -- string representing colors which tikz
          recognizes
        - ``facet_color`` -- string representing colors which tikz
          recognizes
        - ``vertex_color`` -- string representing colors which tikz
          recognizes
        - ``opacity`` -- real number between 0 and 1 giving the opacity of
          the front facets
        - ``axis`` -- boolean draw the axes at the origin or not

        OUTPUT: :class:`LatexExpr` -- containing the TikZ picture

        EXAMPLES::

            sage: # needs sage.plot
            sage: P = Polyhedron(vertices=[[-1, -1, 2], [-1, 2, -1], [2, -1, -1]]); P
            A 2-dimensional polyhedron in ZZ^3 defined as the convex hull of 3 vertices
            sage: Image = P.projection()._tikz_2d_in_3d(view=[0.5, -1, -0.5], angle=55, scale=3,
            ....:                                       edge_color='blue!95!black', facet_color='orange',
            ....:                                       opacity=0.5, vertex_color='yellow', axis=True)
            sage: print('\n'.join(Image.splitlines()[:4]))
            \begin{tikzpicture}%
                [x={(0.644647cm, -0.476559cm)},
                y={(0.192276cm, 0.857859cm)},
                z={(-0.739905cm, -0.192276cm)},
            sage: with open('polytope-tikz3.tex', 'w') as f:    # not tested
            ....:     _ = f.write(Image)

        ::

            sage: p = Polyhedron(vertices=[[1, 0, 0], [0, 1, 0], [0, 0, 1]])
            sage: proj = p.projection()
            sage: Img = proj.tikz([1, 1, 1], 130, axis=True, output_type='LatexExpr')   # needs sage.plot
            sage: print('\n'.join(Img.splitlines()[12:21]))                             # needs sage.plot
            %% with the command: ._tikz_2d_in_3d and parameters:
            %% view = [1, 1, 1]
            %% angle = 130
            %% scale = 1
            %% edge_color = blue!95!black
            %% facet_color = blue!95!black
            %% opacity = 0.8
            %% vertex_color = green
            %% axis = True

        .. NOTE::

            The ``facet_color`` is the filing color of the polytope (polygon).
        """
        from sage.rings.real_double import RDF

        view_vector = vector(RDF, view)
        rot = rotate_arbitrary(view_vector, -(angle/360)*2*pi)
        rotation_matrix = rot[:2].transpose()

        # Creates the nodes, coordinate and tag for every vertex of the polytope.
        # The tag is used to draw the front facets later on.
        dict_drawing = {}
        edges = ''
        for vert in self.points:
            v = self.coords[vert]
            v_vect = str(['%.5f' % i for i in v]).replace('\'', '')
            v_vect = v_vect.replace('[', '(')
            v_vect = v_vect.replace(']', ')')
            tag = '%s' % v_vect
            node = "\\node[%s] at %s     {};\n" % ('vertex', tag)
            coord = '\\coordinate %s at %s;\n' % (tag, tag)
            dict_drawing[vert] = node, coord, tag

        for index1, index2 in self.lines:
            # v1 = self.coords[index1]
            # v2 = self.coords[index2]
            edges += "\\draw[%s] %s -- %s;\n" % ('edge',
                                                 dict_drawing[index1][2],
                                                 dict_drawing[index2][2])

        # Start to write the output
        tikz_pic = ''
        tikz_pic += '\\begin{tikzpicture}%\n'
        tikz_pic += '\t[x={(%fcm, %fcm)},\n' % (RDF(rotation_matrix[0][0]),
                                                RDF(rotation_matrix[0][1]))
        tikz_pic += '\ty={(%fcm, %fcm)},\n' % (RDF(rotation_matrix[1][0]),
                                               RDF(rotation_matrix[1][1]))
        tikz_pic += '\tz={(%fcm, %fcm)},\n' % (RDF(rotation_matrix[2][0]),
                                               RDF(rotation_matrix[2][1]))
        tikz_pic += '\tscale=%f,\n' % scale
        tikz_pic += '\tback/.style={loosely dotted, thin},\n'
        tikz_pic += '\tedge/.style={color=%s, thick},\n' % edge_color
        tikz_pic += '\tfacet/.style={fill=%s,fill opacity=%f},\n' % (facet_color, opacity)
        tikz_pic += '\tvertex/.style={inner sep=1pt,circle,draw=%s!25!black,' % vertex_color
        tikz_pic += 'fill=%s!75!black,thick}]\n%%\n%%\n' % vertex_color

        # Gives the reproduction information
        from sage.env import SAGE_VERSION
        tikz_pic += "%% This TikZ-picture was produced with Sagemath version {}\n".format(SAGE_VERSION)
        tikz_pic += "%% with the command: ._tikz_2d_in_3d and parameters:\n"
        tikz_pic += "%% view = {}\n".format(view)
        tikz_pic += "%% angle = {}\n".format(angle)
        tikz_pic += "%% scale = {}\n".format(scale)
        tikz_pic += "%% edge_color = {}\n".format(edge_color)
        tikz_pic += "%% facet_color = {}\n".format(facet_color)
        tikz_pic += "%% opacity = {}\n".format(opacity)
        tikz_pic += "%% vertex_color = {}\n".format(vertex_color)
        tikz_pic += "%% axis = {}\n%%\n".format(axis)

        # Draws the axes if True
        if axis:
            tikz_pic += '%% Drawing the axes\n'
            tikz_pic += '\\draw[color=black,thick,->] (0,0,0) -- (1,0,0) node[anchor=north east]{$x$};\n'
            tikz_pic += '\\draw[color=black,thick,->] (0,0,0) -- (0,1,0) node[anchor=north west]{$y$};\n'
            tikz_pic += '\\draw[color=black,thick,->] (0,0,0) -- (0,0,1) node[anchor=south]{$z$};\n'

        # Create the coordinate of the vertices:
        tikz_pic += '%% Coordinate of the vertices:\n%%\n'
        for v in dict_drawing:
            tikz_pic += dict_drawing[v][1]

        # Draw the interior by going in a cycle
        vertices = list(self.parent_polyhedron.Vrep_generator())
        tikz_pic += '%%\n%%\n%% Drawing the interior\n%%\n'
        cyclic_vert = cyclic_sort_vertices_2d(list(self.parent_polyhedron.Vrep_generator()))
        cyclic_indices = [vertices.index(v) for v in cyclic_vert]
        tikz_pic += '\\fill[facet] '
        for v in cyclic_indices:
            if v in dict_drawing:
                tikz_pic += '%s -- ' % dict_drawing[v][2]
        tikz_pic += 'cycle {};\n'

        # Draw the edges in the front
        tikz_pic += '%%\n%%\n%% Drawing edges\n%%\n'
        tikz_pic += edges

        # Finally, the vertices in front are drawn on top of everything.
        tikz_pic += '%%\n%%\n%% Drawing the vertices\n%%\n'
        for v in dict_drawing:
            tikz_pic += dict_drawing[v][0]
        tikz_pic += '%%\n%%\n\\end{tikzpicture}'

        return LatexExpr(tikz_pic)

    def _tikz_3d_in_3d(self, view, angle, scale, edge_color,
                       facet_color, opacity, vertex_color, axis):
        r"""
        Return a string ``tikz_pic`` consisting of a tikz picture of ``self``
        according to a projection ``view`` and an angle ``angle``
        obtained via Jmol through the current state property. ``self`` is
        assumed to be a 3-polytope in 3-space.

        INPUT:

        - ``view`` -- list (default: [0,0,1]) representing the rotation axis
        - ``angle`` -- integer angle of rotation in degree from 0 to 360
        - ``scale`` -- integer specifying the scaling of the tikz picture
        - ``edge_color`` -- string representing colors which tikz
          recognizes
        - ``facet_color`` -- string representing colors which tikz
          recognizes
        - ``vertex_color`` -- string representing colors which tikz
          recognizes
        - ``opacity`` -- real number between 0 and 1 giving the opacity of
          the front facets
        - ``axis`` -- boolean draw the axes at the origin or not

        OUTPUT: :class:`LatexExpr` -- containing the TikZ picture.

        EXAMPLES::

            sage: # needs sage.plot sage.rings.number_field
            sage: P = polytopes.small_rhombicuboctahedron()
            sage: Image = P.projection()._tikz_3d_in_3d([3, 7, 5], 100, scale=3,
            ....:                                       edge_color='blue', facet_color='orange',
            ....:                                       opacity=0.5, vertex_color='green', axis=True)
            sage: type(Image)
            <class 'sage.misc.latex.LatexExpr'>
            sage: print('\n'.join(Image.splitlines()[:4]))
            \begin{tikzpicture}%
                [x={(-0.046385cm, 0.837431cm)},
                y={(-0.243536cm, 0.519228cm)},
                z={(0.968782cm, 0.170622cm)},
            sage: with open('polytope-tikz1.tex', 'w') as f:    # not tested
            ....:     _ = f.write(Image)

        ::

            sage: # needs sage.plot
            sage: Associahedron = Polyhedron(vertices=[[1, 0, 1], [1, 0, 0], [1, 1, 0],
            ....:                                      [0, 0, -1], [0, 1, 0], [-1, 0, 0],
            ....:                                      [0, 1, 1], [0, 0, 1], [0, -1, 0]]).polar()
            sage: ImageAsso = Associahedron.projection().tikz([-15, -755, -655], 116, scale=1,
            ....:                                             output_type='LatexExpr')
            sage: print('\n'.join(ImageAsso.splitlines()[12:30]))
            %% with the command: ._tikz_3d_in_3d and parameters:
            %% view = [-15, -755, -655]
            %% angle = 116
            %% scale = 1
            %% edge_color = blue!95!black
            %% facet_color = blue!95!black
            %% opacity = 0.8
            %% vertex_color = green
            %% axis = False
            %%
            %% Coordinate of the vertices:
            %%
            \coordinate (0.00000, 1.00000, -1.00000) at (0.00000, 1.00000, -1.00000);
            \coordinate (1.00000, 1.00000, -1.00000) at (1.00000, 1.00000, -1.00000);
            \coordinate (1.00000, 1.00000, 1.00000) at (1.00000, 1.00000, 1.00000);
            \coordinate (1.00000, -1.00000, 1.00000) at (1.00000, -1.00000, 1.00000);
            \coordinate (1.00000, -1.00000, 0.00000) at (1.00000, -1.00000, 0.00000);
            \coordinate (1.00000, 0.00000, -1.00000) at (1.00000, 0.00000, -1.00000);
        """
        from sage.rings.real_double import RDF

        view_vector = vector(RDF, view)
        rot = rotate_arbitrary(view_vector, -(angle/360)*2*pi)
        rotation_matrix = rot[:2].transpose()
        proj_vector = (rot**(-1))*vector(RDF, [0, 0, 1])

        # First compute the back and front vertices and facets
        facets = self.face_inequalities
        front_facets = []
        back_facets = []
        for index_facet in range(len(facets)):
            A = facets[index_facet].vector()[1:]
            B = facets[index_facet].vector()[0]
            if A*(2000*proj_vector)+B < 0:
                front_facets += [index_facet]
            else:
                back_facets += [index_facet]

        vertices = list(self.parent_polyhedron.Vrep_generator())
        front_vertices = []
        for index_facet in front_facets:
            A = facets[index_facet].vector()[1:]
            B = facets[index_facet].vector()[0]
            for v in self.points:
                if A*self.coords[v]+B < 0.0005 and v not in front_vertices:
                    front_vertices += [v]

        back_vertices = []
        for index_facet in back_facets:
            A = facets[index_facet].vector()[1:]
            B = facets[index_facet].vector()[0]
            for v in self.points:
                if A * self.coords[v] + B < 0.0005 and v not in back_vertices:
                    back_vertices += [v]

        # Creates the nodes, coordinate and tag for every vertex of the polytope.
        # The tag is used to draw the front facets later on.
        dict_drawing = {}
        back_part = ''
        front_part = ''

        for vert in self.points:
            v = self.coords[vert]
            v_vect = str(['%.5f' % i for i in v]).replace('\'', '')
            v_vect = v_vect.replace('[', '(')
            v_vect = v_vect.replace(']', ')')
            tag = '%s' % v_vect
            node = "\\node[%s] at %s     {};\n" % ('vertex', tag)
            coord = '\\coordinate %s at %s;\n' % (tag, tag)
            dict_drawing[vert] = node, coord, tag

        # Separate the edges between back and front

        for index1, index2 in self.lines:
            # v1 = self.coords[index1]
            # v2 = self.coords[index2]

            H_v1 = set(self.parent_polyhedron.Vrepresentation(index1).incident())
            H_v2 = set(self.parent_polyhedron.Vrepresentation(index2).incident())
            H_v12 = [h for h in H_v1.intersection(H_v2) if facets.index(h) in back_facets]

            # The back edge has to be between two vertices in the Back
            # AND such that the 2 facets touching them are in the Back
            if index1 in back_vertices and index2 in back_vertices and len(H_v12) == 2:
                back_part += "\\draw[%s,back] %s -- %s;\n" % ('edge',
                                                              dict_drawing[index1][2],
                                                              dict_drawing[index2][2])
            else:
                front_part += "\\draw[%s] %s -- %s;\n" % ('edge',
                                                          dict_drawing[index1][2],
                                                          dict_drawing[index2][2])

        # Start to write the output
        tikz_pic = ''
        tikz_pic += '\\begin{tikzpicture}%\n'
        tikz_pic += '\t[x={(%fcm, %fcm)},\n' % (RDF(rotation_matrix[0][0]),
                                                RDF(rotation_matrix[0][1]))
        tikz_pic += '\ty={(%fcm, %fcm)},\n' % (RDF(rotation_matrix[1][0]),
                                               RDF(rotation_matrix[1][1]))
        tikz_pic += '\tz={(%fcm, %fcm)},\n' % (RDF(rotation_matrix[2][0]),
                                               RDF(rotation_matrix[2][1]))
        tikz_pic += '\tscale=%f,\n' % scale
        tikz_pic += '\tback/.style={loosely dotted, thin},\n'
        tikz_pic += '\tedge/.style={color=%s, thick},\n' % edge_color
        tikz_pic += '\tfacet/.style={fill=%s,fill opacity=%f},\n' % (facet_color, opacity)
        tikz_pic += '\tvertex/.style={inner sep=1pt,circle,draw=%s!25!black,' % vertex_color
        tikz_pic += 'fill=%s!75!black,thick}]\n%%\n%%\n' % vertex_color

        # Gives the reproduction information
        from sage.env import SAGE_VERSION
        tikz_pic += "%% This TikZ-picture was produced with Sagemath version {}\n".format(SAGE_VERSION)
        tikz_pic += "%% with the command: ._tikz_3d_in_3d and parameters:\n"
        tikz_pic += "%% view = {}\n".format(view)
        tikz_pic += "%% angle = {}\n".format(angle)
        tikz_pic += "%% scale = {}\n".format(scale)
        tikz_pic += "%% edge_color = {}\n".format(edge_color)
        tikz_pic += "%% facet_color = {}\n".format(facet_color)
        tikz_pic += "%% opacity = {}\n".format(opacity)
        tikz_pic += "%% vertex_color = {}\n".format(vertex_color)
        tikz_pic += "%% axis = {}\n%%\n".format(axis)

        # Draws the axes if True
        if axis:
            tikz_pic += '%% Drawing the axes\n'
            tikz_pic += '\\draw[color=black,thick,->] (0,0,0) -- (1,0,0) node[anchor=north east]{$x$};\n'
            tikz_pic += '\\draw[color=black,thick,->] (0,0,0) -- (0,1,0) node[anchor=north west]{$y$};\n'
            tikz_pic += '\\draw[color=black,thick,->] (0,0,0) -- (0,0,1) node[anchor=south]{$z$};\n'

        # Create the coordinate of the vertices
        tikz_pic += '%% Coordinate of the vertices:\n%%\n'
        for v in dict_drawing:
            tikz_pic += dict_drawing[v][1]

        # Draw the edges in the back
        tikz_pic += '%%\n%%\n%% Drawing edges in the back\n%%\n'
        tikz_pic += back_part

        # Draw the vertices on top of the back-edges
        tikz_pic += '%%\n%%\n%% Drawing vertices in the back\n%%\n'
        for v in back_vertices:
            if v not in front_vertices and v in dict_drawing:
                tikz_pic += dict_drawing[v][0]

        # Draw the facets in the front by going in cycles for every facet.
        tikz_pic += '%%\n%%\n%% Drawing the facets\n%%\n'
        vertex_to_index = {v: i for i, v in enumerate(vertices)}
        for index_facet in front_facets:
            cyclic_vert = cyclic_sort_vertices_2d(list(facets[index_facet].incident()))
            cyclic_indices = [vertex_to_index[v] for v in cyclic_vert]
            tikz_pic += '\\fill[facet] '
            for v in cyclic_indices:
                if v in dict_drawing:
                    tikz_pic += '%s -- ' % dict_drawing[v][2]
            tikz_pic += 'cycle {};\n'

        # Draw the edges in the front
        tikz_pic += '%%\n%%\n%% Drawing edges in the front\n%%\n'
        tikz_pic += front_part

        # Finally, the vertices in front are drawn on top of everything.
        tikz_pic += '%%\n%%\n%% Drawing the vertices in the front\n%%\n'
        for v in self.points:
            if v in front_vertices:
                if v in dict_drawing:
                    tikz_pic += dict_drawing[v][0]
        tikz_pic += '%%\n%%\n\\end{tikzpicture}'
        return LatexExpr(tikz_pic)



# sage_setup: distribution = sagemath-polyhedra
"""
A triangulation

In Sage, the
:class:`~sage.geometry.triangulation.point_configuration.PointConfiguration`
and :class:`Triangulation` satisfy a parent/element relationship. In
particular, each triangulation refers back to its point
configuration. If you want to triangulate a point configuration, you
should construct a point configuration first and then use one of its
methods to triangulate it according to your requirements. You should
never have to construct a :class:`Triangulation` object directly.

EXAMPLES:

First, we select the internal implementation for enumerating
triangulations::

    sage: PointConfiguration.set_engine('internal')   # to make doctests independent of TOPCOM

Here is a simple example of how to triangulate a point configuration::

    sage: p = [[0,-1,-1],[0,0,1],[0,1,0], [1,-1,-1],[1,0,1],[1,1,0]]
    sage: points = PointConfiguration(p)
    sage: triang = points.triangulate();  triang
    (<0,1,2,5>, <0,1,3,5>, <1,3,4,5>)
    sage: triang.plot(axes=False)                                                       # needs sage.plot
    Graphics3d Object

See :mod:`sage.geometry.triangulation.point_configuration` for more details.
"""

#*****************************************************************************
#       Copyright (C) 2010 Volker Braun <vbraun.name@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from sage.structure.richcmp import richcmp
from sage.structure.element import Element
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.modules.free_module_element import vector
from sage.misc.cachefunc import cached_method
from sage.sets.set import Set


########################################################################
def triangulation_render_2d(triangulation, **kwds):
    r"""
    Return a graphical representation of a 2-d triangulation.

    INPUT:

    - ``triangulation`` -- a :class:`Triangulation`

    - ``**kwds`` -- keywords that are passed on to the graphics primitives

    OUTPUT: a 2-d graphics object

    EXAMPLES::

        sage: points = PointConfiguration([[0,0],[0,1],[1,0],[1,1],[-1,-1]])
        sage: triang = points.triangulate()
        sage: triang.plot(axes=False, aspect_ratio=1)   # indirect doctest              # needs sage.plot
        Graphics object consisting of 12 graphics primitives
    """
    from sage.plot.all import point2d, line2d, polygon2d
    points = [point.reduced_affine() for point in triangulation.point_configuration()]
    coord = [ [p[0], p[1]] for p in points ]
    plot_points = sum([ point2d(p,
                                zorder=2, pointsize=10, **kwds)
                        for p in coord ])

    tmp_lines = []
    for t in triangulation:
        if len(t) >= 2:
            tmp_lines.append([t[0], t[1]])
        if len(t) >= 3:
            tmp_lines.append([t[0], t[2]])
            tmp_lines.append([t[1], t[2]])
    all_lines = []
    interior_lines = []
    for l in tmp_lines:
        if l not in all_lines:
            all_lines.append(l)
        else:
            interior_lines.append(l)
    exterior_lines = [l for l in all_lines if l not in interior_lines]

    plot_interior_lines = sum([ line2d([ coord[l[0]], coord[l[1]] ],
                                       zorder=1, rgbcolor=(0,1,0), **kwds)
                                for l in interior_lines ])
    plot_exterior_lines = sum([ line2d([ coord[l[0]], coord[l[1]] ],
                                       zorder=1, rgbcolor=(0,0,1), **kwds)
                                for l in exterior_lines ])

    plot_triangs = sum([ polygon2d([coord[t[0]], coord[t[1]], coord[t[2]]],
                                   zorder=0, rgbcolor=(0.8, 1, 0.8), **kwds)
                         for t in triangulation if len(t) >= 3 ])

    return \
        plot_points + \
        plot_interior_lines + plot_exterior_lines + \
        plot_triangs


def triangulation_render_3d(triangulation, **kwds):
    r"""
    Return a graphical representation of a 3-d triangulation.

    INPUT:

    - ``triangulation`` -- a :class:`Triangulation`

    - ``**kwds`` -- keywords that are  passed on to the graphics primitives

    OUTPUT: a 3-d graphics object

    EXAMPLES::

        sage: p = [[0,-1,-1],[0,0,1],[0,1,0], [1,-1,-1],[1,0,1],[1,1,0]]
        sage: points = PointConfiguration(p)
        sage: triang = points.triangulate()
        sage: triang.plot(axes=False)     # indirect doctest                            # needs sage.plot
        Graphics3d Object
    """
    from sage.plot.plot3d.all import point3d, line3d, polygon3d
    points = [ point.reduced_affine() for point in triangulation.point_configuration() ]
    coord = [ [p[0], p[1], p[2] ] for p in points ]
    plot_points = sum([ point3d(p, size=15,
                                **kwds)
                        for p in coord ])

    tmp_lines = []
    for t in triangulation:
        if len(t) >= 2:
            tmp_lines.append([t[0], t[1]])
        if len(t) >= 3:
            tmp_lines.append([t[0], t[2]])
            tmp_lines.append([t[1], t[2]])
        if len(t) >= 4:
            tmp_lines.append([t[0], t[3]])
            tmp_lines.append([t[1], t[3]])
            tmp_lines.append([t[2], t[3]])
    all_lines = []
    interior_lines = []
    for l in tmp_lines:
        if l not in all_lines:
            all_lines.append(l)
        else:
            interior_lines.append(l)
    exterior_lines = [l for l in all_lines if l not in interior_lines]

    from sage.plot.plot3d.texture import Texture
    line_int = Texture(color='darkblue', ambient=1, diffuse=0)
    line_ext = Texture(color='green', ambient=1, diffuse=0)
    triang_int = Texture(opacity=0.3, specular=0, shininess=0, diffuse=0, ambient=1, color='yellow')
    triang_ext = Texture(opacity=0.6, specular=0, shininess=0, diffuse=0, ambient=1, color='green')

    plot_interior_lines = sum([ line3d([ coord[l[0]], coord[l[1]] ],
                                       thickness=2, texture=line_int, **kwds)
                                for l in interior_lines ])
    plot_exterior_lines = sum([ line3d([ coord[l[0]], coord[l[1]] ],
                                       thickness=3, texture=line_ext, **kwds)
                                for l in exterior_lines ])

    tmp_triangs = []
    for t in triangulation:
        if len(t) >= 3:
            tmp_triangs.append([t[0], t[1], t[2]])
        if len(t) >= 4:
            tmp_triangs.append([t[0], t[1], t[3]])
            tmp_triangs.append([t[0], t[2], t[3]])
            tmp_triangs.append([t[1], t[2], t[3]])
    all_triangs = []
    interior_triangs = []
    for l in tmp_triangs:
        if l not in all_triangs:
            all_triangs.append(l)
        else:
            interior_triangs.append(l)
    exterior_triangs = [l for l in all_triangs if l not in interior_triangs]

    plot_interior_triangs = \
        sum([polygon3d([coord[t[0]], coord[t[1]], coord[t[2]]],
                       texture=triang_int, **kwds)
             for t in interior_triangs])
    plot_exterior_triangs = \
        sum([polygon3d([coord[t[0]], coord[t[1]], coord[t[2]]],
                       texture=triang_ext, **kwds)
              for t in exterior_triangs])

    return plot_points + \
        plot_interior_lines + plot_exterior_lines + \
        plot_interior_triangs + plot_exterior_triangs


########################################################################
class Triangulation(Element):
    """
    A triangulation of a
    :class:`~sage.geometry.triangulation.point_configuration.PointConfiguration`.

    .. WARNING::

        You should never create :class:`Triangulation` objects
        manually. See
        :meth:`~sage.geometry.triangulation.point_configuration.PointConfiguration.triangulate`
        and
        :meth:`~sage.geometry.triangulation.point_configuration.PointConfiguration.triangulations`
        to triangulate point configurations.
    """
    def __init__(self, triangulation, parent, check=True):
        """
        The constructor of a ``Triangulation`` object.

        Note that an internal reference to the underlying ``PointConfiguration`` is
        kept.

        INPUT:

        - ``parent`` -- a
          :class:`~sage.geometry.triangulation.point_configuration.PointConfiguration`

        - ``triangulation`` -- an iterable of integers or an iterable of
          iterables (e.g. a list of lists), specifying the maximal simplices
          of the triangulation. In the first case, each integer specifies a simplex
          by the correspondence :meth:`PointConfiguration.simplex_to_int`. In the second
          case, a simplex is specified by listing the indices of the included points.

        - ``check`` -- boolean; whether to perform checks that the
          triangulation is, indeed, a triangulation of the point
          configuration

        NOTE:

        Passing ``check=False`` allows you to create triangulations of
        subsets of the points of the configuration, see
        :meth:`~sage.geometry.triangulation.point_configuration.PointConfiguration.bistellar_flips`.

        EXAMPLES::

            sage: p = [[0,1],[0,0],[1,0]]
            sage: points = PointConfiguration(p)
            sage: from sage.geometry.triangulation.point_configuration import Triangulation
            sage: Triangulation([(0,1,2)], points)
            (<0,1,2>)
            sage: Triangulation([1], points)
            (<0,1,2>)
        """
        Element.__init__(self, parent=parent)
        self._point_configuration = parent

        try:
            triangulation = tuple(sorted( tuple(sorted(t)) for t in triangulation))
        except TypeError:
            triangulation = tuple( self.point_configuration().int_to_simplex(i)
                                   for i in triangulation )
        assert not check or all( len(t) == self.point_configuration().dim()+1
                                 for t in triangulation)
        self._triangulation = triangulation

    def point_configuration(self):
        """
        Return the point configuration underlying the triangulation.

        EXAMPLES::

            sage: pconfig = PointConfiguration([[0,0],[0,1],[1,0]])
            sage: pconfig
            A point configuration in affine 2-space over Integer Ring
            consisting of 3 points. The triangulations of this point
            configuration are assumed to be connected, not necessarily
            fine, not necessarily regular.
            sage: triangulation = pconfig.triangulate()
            sage: triangulation
            (<0,1,2>)
            sage: triangulation.point_configuration()
            A point configuration in affine 2-space over Integer Ring
            consisting of 3 points. The triangulations of this point
            configuration are assumed to be connected, not necessarily
            fine, not necessarily regular.
            sage: pconfig == triangulation.point_configuration()
            True
        """
        return self._point_configuration

    def _richcmp_(self, right, op):
        r"""
        Compare ``self`` and ``right``.

        INPUT:

        - ``right`` -- a triangulation

        TESTS::

            sage: pc = PointConfiguration([[0,0],[0,1],[1,0]])
            sage: t1 = pc.triangulate()
            sage: from sage.geometry.triangulation.point_configuration import Triangulation
            sage: t2 = Triangulation([[2,1,0]], pc)
            sage: t1 is t2
            False
            sage: t1 == t2    # indirect doctest
            True
            sage: t1 != Triangulation(((0,1),(1,2)), pc, check=False)
            True
        """
        return richcmp(self._triangulation, right._triangulation, op)

    def __iter__(self):
        """
        Iterate through the simplices of the triangulation.

        EXAMPLES::

            sage: PointConfiguration.set_engine('internal')   # to make doctests independent of TOPCOM
            sage: pc = PointConfiguration([[0,0],[0,1],[1,0],[1,1],[-1,-1]])
            sage: triangulation = pc.triangulate()
            sage: iter = triangulation.__iter__()
            sage: next(iter)
            (1, 3, 4)
            sage: next(iter)
            (2, 3, 4)
            sage: next(iter)
            Traceback (most recent call last):
            ...
            StopIteration
        """
        yield from self._triangulation

    def __getitem__(self, i):
        """
        Access the point indices of the `i`-th simplex of the triangulation.

        INPUT:

        - ``i`` -- integer; the index of a simplex

        OUTPUT:

        A tuple of integers. The vertex indices of the `i`-th simplex.

        EXAMPLES::

            sage: PointConfiguration.set_engine('internal')   # to make doctests independent of TOPCOM
            sage: pc = PointConfiguration([[0,0],[0,1],[1,0],[1,1],[-1,-1]])
            sage: triangulation = pc.triangulate()
            sage: triangulation[1]
            (2, 3, 4)
        """
        return self._triangulation[i]

    def __len__(self):
        """
        Return the length of the triangulation.

        TESTS::

            sage: PointConfiguration.set_engine('internal')   # to make doctests independent of TOPCOM
            sage: pc = PointConfiguration([[0,0],[0,1],[1,0],[1,1],[-1,-1]])
            sage: triangulation = next(pc.triangulations())
            sage: triangulation.__len__()
            2
            sage: len(triangulation)    # equivalent
            2
        """
        return len(self._triangulation)

    def _repr_(self):
        r"""
        Return a string representation.

        TESTS::

            sage: PointConfiguration.set_engine('internal')   # to make doctests independent of TOPCOM
            sage: pc = PointConfiguration([[0,0],[0,1],[1,0],[1,1],[-1,-1],[2,2]])
            sage: t = pc.triangulations()
            sage: next(t)._repr_()
            '(<1,4,5>, <2,4,5>)'
        """
        #s = 'A triangulation'
        #s += ' in QQ^'+str(self.point_configuration().ambient_dim())
        #s += ' consisting of '+str(len(self))+' simplices.'
        s = '('
        s += ', '.join([ '<'+','.join(map(str,t))+'>' for t in self._triangulation])
        s += ')'
        return s

    def plot(self, **kwds):
        r"""
        Produce a graphical representation of the triangulation.

        EXAMPLES::

            sage: p = PointConfiguration([[0,0],[0,1],[1,0],[1,1],[-1,-1]])
            sage: triangulation = p.triangulate()
            sage: triangulation
            (<1,3,4>, <2,3,4>)
            sage: triangulation.plot(axes=False)                                        # needs sage.plot
            Graphics object consisting of 12 graphics primitives
        """
        dim = self.point_configuration().dim()

        if dim == 2:
            return triangulation_render_2d(self, **kwds)

        if dim == 3:
            return triangulation_render_3d(self, **kwds)

        raise NotImplementedError('Plotting '+str(dim)+'-dimensional triangulations not implemented!')

    def gkz_phi(self):
        r"""
        Calculate the GKZ phi vector of the triangulation.

        The phi vector is a vector of length equals to the number of
        points in the point configuration. For a fixed triangulation
        `T`, the entry corresponding to the `i`-th point `p_i` is

        .. MATH::

            \phi_T(p_i) = \sum_{t\in T, t\owns p_i} Vol(t)

        that is, the total volume of all simplices containing `p_i`.
        See also [GKZ1994]_ page 220 equation 1.4.

        OUTPUT: the phi vector of self

        EXAMPLES::

            sage: p = PointConfiguration([[0,0],[1,0],[2,1],[1,2],[0,1]])
            sage: p.triangulate().gkz_phi()
            (3, 1, 5, 2, 4)
            sage: p.lexicographic_triangulation().gkz_phi()
            (1, 3, 4, 2, 5)
        """
        vec = vector(ZZ, self.point_configuration().n_points())
        for simplex in self:
            vol = self.point_configuration().volume(simplex)
            for i in simplex:
                vec[i] = vec[i] + vol
        return vec

    def enumerate_simplices(self):
        r"""
        Return the enumerated simplices.

        OUTPUT: a tuple of integers that uniquely specifies the triangulation

        EXAMPLES::

            sage: pc = PointConfiguration(matrix([
            ....:    [ 0, 0, 0, 0, 0, 2, 4,-1, 1, 1, 0, 0, 1, 0],
            ....:    [ 0, 0, 0, 1, 0, 0,-1, 0, 0, 0, 0, 0, 0, 0],
            ....:    [ 0, 2, 0, 0, 0, 0,-1, 0, 1, 0, 1, 0, 0, 1],
            ....:    [ 0, 1, 1, 0, 0, 1, 0,-2, 1, 0, 0,-1, 1, 1],
            ....:    [ 0, 0, 0, 0, 1, 0,-1, 0, 0, 0, 0, 0, 0, 0]
            ....: ]).columns())
            sage: triangulation = pc.lexicographic_triangulation()
            sage: triangulation.enumerate_simplices()
            (1678, 1688, 1769, 1779, 1895, 1905, 2112, 2143, 2234, 2360, 2555, 2580,
             2610, 2626, 2650, 2652, 2654, 2661, 2663, 2667, 2685, 2755, 2757, 2759,
             2766, 2768, 2772, 2811, 2881, 2883, 2885, 2892, 2894, 2898)

        You can recreate the triangulation from this list by passing
        it to the constructor::

            sage: from sage.geometry.triangulation.point_configuration import Triangulation
            sage: Triangulation([1678, 1688, 1769, 1779, 1895, 1905, 2112, 2143,
            ....:  2234, 2360, 2555, 2580, 2610, 2626, 2650, 2652, 2654, 2661, 2663,
            ....:  2667, 2685, 2755, 2757, 2759, 2766, 2768, 2772, 2811, 2881, 2883,
            ....:  2885, 2892, 2894, 2898], pc)
            (<1,3,4,7,10,13>, <1,3,4,8,10,13>, <1,3,6,7,10,13>, <1,3,6,8,10,13>,
             <1,4,6,7,10,13>, <1,4,6,8,10,13>, <2,3,4,6,7,12>, <2,3,4,7,12,13>,
             <2,3,6,7,12,13>, <2,4,6,7,12,13>, <3,4,5,6,9,12>, <3,4,5,8,9,12>,
             <3,4,6,7,11,12>, <3,4,6,9,11,12>, <3,4,7,10,11,13>, <3,4,7,11,12,13>,
             <3,4,8,9,10,12>, <3,4,8,10,12,13>, <3,4,9,10,11,12>, <3,4,10,11,12,13>,
             <3,5,6,8,9,12>, <3,6,7,10,11,13>, <3,6,7,11,12,13>, <3,6,8,9,10,12>,
             <3,6,8,10,12,13>, <3,6,9,10,11,12>, <3,6,10,11,12,13>, <4,5,6,8,9,12>,
             <4,6,7,10,11,13>, <4,6,7,11,12,13>, <4,6,8,9,10,12>, <4,6,8,10,12,13>,
             <4,6,9,10,11,12>, <4,6,10,11,12,13>)
        """
        pc = self._point_configuration
        return tuple( pc.simplex_to_int(t) for t in self )

    def fan(self, origin=None):
        r"""
        Construct the fan of cones over the simplices of the triangulation.

        INPUT:

        - ``origin`` -- ``None`` (default) or coordinates of a
          point. The common apex of all cones of the fan. If ``None``,
          the triangulation must be a star triangulation and the
          distinguished central point is used as the origin.

        OUTPUT:

        A :class:`~sage.geometry.fan.RationalPolyhedralFan`. The
        coordinates of the points are shifted so that the apex of the
        fan is the origin of the coordinate system.

        .. NOTE:: If the set of cones over the simplices is not a fan, a
            suitable exception is raised.

        EXAMPLES::

            sage: pc = PointConfiguration([(0,0), (1,0), (0,1), (-1,-1)], star=0, fine=True)
            sage: triangulation = pc.triangulate()
            sage: fan = triangulation.fan(); fan
            Rational polyhedral fan in 2-d lattice N
            sage: fan.is_equivalent(toric_varieties.P2().fan())                         # needs palp sage.graphs
            True

        Toric diagrams (the `\ZZ_5` hyperconifold)::

            sage: vertices=[(0, 1, 0), (0, 3, 1), (0, 2, 3), (0, 0, 2)]
            sage: interior=[(0, 1, 1), (0, 1, 2), (0, 2, 1), (0, 2, 2)]
            sage: points = vertices + interior
            sage: pc = PointConfiguration(points, fine=True)
            sage: triangulation = pc.triangulate()
            sage: fan = triangulation.fan((-1,0,0)); fan
            Rational polyhedral fan in 3-d lattice N
            sage: fan.rays()
            N(1, 1, 0),
            N(1, 3, 1),
            N(1, 2, 3),
            N(1, 0, 2),
            N(1, 1, 1),
            N(1, 1, 2),
            N(1, 2, 1),
            N(1, 2, 2)
            in 3-d lattice N
        """
        from sage.geometry.fan import Fan
        if origin is None:
            origin = self.point_configuration().star_center()
        R = self.base_ring()
        origin = vector(R, origin)
        points = self.point_configuration().points()
        return Fan(self, (vector(R, p) - origin for p in points))

    @cached_method
    def simplicial_complex(self):
        r"""
        Return ``self`` as an (abstract) simplicial complex.

        OUTPUT: a :class:`~sage.topology.simplicial_complex.SimplicialComplex`

        EXAMPLES::

            sage: p = polytopes.cuboctahedron()
            sage: sc = p.triangulate(engine='internal').simplicial_complex(); sc        # needs sage.graphs
            Simplicial complex with 12 vertices and 16 facets

        Any convex set is contractable, so its reduced homology groups vanish::

            sage: sc.homology()                                                         # needs sage.graphs
            {0: 0, 1: 0, 2: 0, 3: 0}
        """
        from sage.topology.simplicial_complex import SimplicialComplex
        return SimplicialComplex(self)

    @cached_method
    def _boundary_simplex_dictionary(self):
        """
        Return facets and the simplices they bound.

        TESTS::

            sage: triangulation = polytopes.hypercube(2).triangulate(engine='internal')
            sage: triangulation._boundary_simplex_dictionary()
            {(0, 1): ((0, 1, 3),),
             (0, 3): ((0, 1, 3),),
             (1, 2): ((1, 2, 3),),
             (1, 3): ((0, 1, 3), (1, 2, 3)),
             (2, 3): ((1, 2, 3),)}

            sage: triangulation = polytopes.cube().triangulate(engine='internal')
            sage: triangulation._boundary_simplex_dictionary()
            {(0, 1, 2): ((0, 1, 2, 7),),
             (0, 1, 5): ((0, 1, 5, 7),),
             (0, 1, 7): ((0, 1, 2, 7), (0, 1, 5, 7)),
             (0, 2, 3): ((0, 2, 3, 7),),
             (0, 2, 7): ((0, 1, 2, 7), (0, 2, 3, 7)),
             (0, 3, 4): ((0, 3, 4, 7),),
             (0, 3, 7): ((0, 2, 3, 7), (0, 3, 4, 7)),
             (0, 4, 5): ((0, 4, 5, 7),),
             (0, 4, 7): ((0, 3, 4, 7), (0, 4, 5, 7)),
             (0, 5, 7): ((0, 1, 5, 7), (0, 4, 5, 7)),
             (1, 2, 7): ((0, 1, 2, 7),),
             (1, 5, 6): ((1, 5, 6, 7),),
             (1, 5, 7): ((0, 1, 5, 7), (1, 5, 6, 7)),
             (1, 6, 7): ((1, 5, 6, 7),),
             (2, 3, 7): ((0, 2, 3, 7),),
             (3, 4, 7): ((0, 3, 4, 7),),
             (4, 5, 7): ((0, 4, 5, 7),),
             (5, 6, 7): ((1, 5, 6, 7),)}
        """
        result = dict()
        for simplex in self:
            for i in range(len(simplex)):
                facet = simplex[:i] + simplex[i+1:]
                result[facet] = result.get(facet, tuple()) + (simplex,)
        return result

    @cached_method
    def boundary(self):
        """
        Return the boundary of the triangulation.

        OUTPUT:

        The outward-facing boundary simplices (of dimension `d-1`) of
        the `d`-dimensional triangulation as a set. Each boundary is
        returned by a tuple of point indices.

        EXAMPLES::

            sage: triangulation = polytopes.cube().triangulate(engine='internal')
            sage: triangulation
            (<0,1,2,7>, <0,1,5,7>, <0,2,3,7>, <0,3,4,7>, <0,4,5,7>, <1,5,6,7>)
            sage: triangulation.boundary()
            frozenset({(0, 1, 2),
                       (0, 1, 5),
                       (0, 2, 3),
                       (0, 3, 4),
                       (0, 4, 5),
                       (1, 2, 7),
                       (1, 5, 6),
                       (1, 6, 7),
                       (2, 3, 7),
                       (3, 4, 7),
                       (4, 5, 7),
                       (5, 6, 7)})
            sage: triangulation.interior_facets()
            frozenset({(0, 1, 7), (0, 2, 7), (0, 3, 7), (0, 4, 7), (0, 5, 7), (1, 5, 7)})
        """
        return frozenset(facet for facet, bounded_simplices
                         in self._boundary_simplex_dictionary().items()
                         if len(bounded_simplices) == 1)

    @cached_method
    def boundary_simplicial_complex(self):
        r"""
        Return the boundary of ``self`` as an (abstract) simplicial complex.

        OUTPUT: a :class:`~sage.topology.simplicial_complex.SimplicialComplex`

        EXAMPLES::

            sage: p = polytopes.cuboctahedron()
            sage: triangulation = p.triangulate(engine='internal')
            sage: bd_sc = triangulation.boundary_simplicial_complex(); bd_sc            # needs sage.graphs
            Simplicial complex with 12 vertices and 20 facets

        The boundary of every convex set is a topological sphere, so it has
        spherical homology::

            sage: bd_sc.homology()                                                      # needs sage.graphs
            {0: 0, 1: 0, 2: Z}

        It is a subcomplex of ``self`` as a :meth:`simplicial_complex`::

            sage: sc = triangulation.simplicial_complex()                               # needs sage.graphs
            sage: all(f in sc for f in bd_sc.maximal_faces())                           # needs sage.graphs
            True
        """
        from sage.topology.simplicial_complex import SimplicialComplex
        return SimplicialComplex(self.boundary(), maximality_check=False)

    @cached_method
    def interior_facets(self):
        """
        Return the interior facets of the triangulation.

        OUTPUT:

        The inward-facing boundary simplices (of dimension `d-1`) of
        the `d`-dimensional triangulation as a set. Each boundary is
        returned by a tuple of point indices.

        EXAMPLES::

            sage: triangulation = polytopes.cube().triangulate(engine='internal')
            sage: triangulation
            (<0,1,2,7>, <0,1,5,7>, <0,2,3,7>, <0,3,4,7>, <0,4,5,7>, <1,5,6,7>)
            sage: triangulation.boundary()
            frozenset({(0, 1, 2),
                       (0, 1, 5),
                       (0, 2, 3),
                       (0, 3, 4),
                       (0, 4, 5),
                       (1, 2, 7),
                       (1, 5, 6),
                       (1, 6, 7),
                       (2, 3, 7),
                       (3, 4, 7),
                       (4, 5, 7),
                       (5, 6, 7)})
            sage: triangulation.interior_facets()
            frozenset({(0, 1, 7), (0, 2, 7), (0, 3, 7), (0, 4, 7), (0, 5, 7), (1, 5, 7)})
        """
        return frozenset(facet for facet, bounded_simplices
                         in self._boundary_simplex_dictionary().items()
                         if len(bounded_simplices) == 2)

    def polyhedral_complex(self, **kwds):
        """
        Return ``self`` as a :class:`~sage.geometry.polyhedral_complex.PolyhedralComplex`.

        OUTPUT:

        A :class:`~sage.geometry.polyhedral_complex.PolyhedralComplex` whose maximal cells
        are the simplices of the triangulation.

        EXAMPLES::

            sage: P = polytopes.cube()
            sage: pc = PointConfiguration(P.vertices())
            sage: T = pc.placing_triangulation(); T
            (<0,1,2,7>, <0,1,5,7>, <0,2,3,7>, <0,3,4,7>, <0,4,5,7>, <1,5,6,7>)
            sage: C = T.polyhedral_complex(); C                                         # needs sage.graphs
            Polyhedral complex with 6 maximal cells
            sage: [P.vertices_list() for P in C.maximal_cells_sorted()]                 # needs sage.graphs
            [[[-1, -1, -1], [-1, -1, 1], [-1, 1, 1], [1, -1, -1]],
             [[-1, -1, -1], [-1, 1, -1], [-1, 1, 1], [1, 1, -1]],
             [[-1, -1, -1], [-1, 1, 1], [1, -1, -1], [1, 1, -1]],
             [[-1, -1, 1], [-1, 1, 1], [1, -1, -1], [1, -1, 1]],
             [[-1, 1, 1], [1, -1, -1], [1, -1, 1], [1, 1, 1]],
             [[-1, 1, 1], [1, -1, -1], [1, 1, -1], [1, 1, 1]]]
        """
        from sage.geometry.polyhedral_complex import PolyhedralComplex
        from sage.geometry.polyhedron.constructor import Polyhedron
        ambient_dim = self.point_configuration().ambient_dim()
        points = self.point_configuration().points()
        return PolyhedralComplex([Polyhedron(vertices=[points[i] for i in simplex])
                                  for simplex in self],
                                 ambient_dim=ambient_dim,
                                 maximality_check=False,
                                 face_to_face_check=False,
                                 **kwds)

    def boundary_polyhedral_complex(self, **kwds):
        r"""
        Return the boundary of ``self`` as a :class:`~sage.geometry.polyhedral_complex.PolyhedralComplex`.

        OUTPUT:

        A :class:`~sage.geometry.polyhedral_complex.PolyhedralComplex` whose maximal cells
        are the simplices of the boundary of ``self``.

        EXAMPLES::

            sage: P = polytopes.cube()
            sage: pc = PointConfiguration(P.vertices())
            sage: T = pc.placing_triangulation(); T
            (<0,1,2,7>, <0,1,5,7>, <0,2,3,7>, <0,3,4,7>, <0,4,5,7>, <1,5,6,7>)
            sage: bd_C = T.boundary_polyhedral_complex(); bd_C                          # needs sage.graphs
            Polyhedral complex with 12 maximal cells
            sage: [P.vertices_list() for P in bd_C.maximal_cells_sorted()]              # needs sage.graphs
            [[[-1, -1, -1], [-1, -1, 1], [-1, 1, 1]],
             [[-1, -1, -1], [-1, -1, 1], [1, -1, -1]],
             [[-1, -1, -1], [-1, 1, -1], [-1, 1, 1]],
             [[-1, -1, -1], [-1, 1, -1], [1, 1, -1]],
             [[-1, -1, -1], [1, -1, -1], [1, 1, -1]],
             [[-1, -1, 1], [-1, 1, 1], [1, -1, 1]],
             [[-1, -1, 1], [1, -1, -1], [1, -1, 1]],
             [[-1, 1, -1], [-1, 1, 1], [1, 1, -1]],
             [[-1, 1, 1], [1, -1, 1], [1, 1, 1]],
             [[-1, 1, 1], [1, 1, -1], [1, 1, 1]],
             [[1, -1, -1], [1, -1, 1], [1, 1, 1]],
             [[1, -1, -1], [1, 1, -1], [1, 1, 1]]]

        It is a subcomplex of ``self`` as a :meth:`polyhedral_complex`::

            sage: C = T.polyhedral_complex()                                            # needs sage.graphs
            sage: bd_C.is_subcomplex(C)                                                 # needs sage.graphs
            True
        """
        from sage.geometry.polyhedral_complex import PolyhedralComplex
        from sage.geometry.polyhedron.constructor import Polyhedron
        ambient_dim = self.point_configuration().ambient_dim()
        points = self.point_configuration().points()
        return PolyhedralComplex([Polyhedron(vertices=[points[i] for i in simplex])
                                  for simplex in self.boundary()],
                                 ambient_dim=ambient_dim,
                                 maximality_check=False,
                                 face_to_face_check=False,
                                 **kwds)

    @cached_method
    def normal_cone(self):
        r"""
        Return the (closure of the) normal cone of the triangulation.

        Recall that a regular triangulation is one that equals the
        "crease lines" of a convex piecewise-linear function. This
        support function is not unique, for example, you can scale it
        by a positive constant. The set of all piecewise-linear
        functions with fixed creases forms an open cone. This cone can
        be interpreted as the cone of normal vectors at a point of the
        secondary polytope, which is why we call it normal cone. See
        [GKZ1994]_ Section 7.1 for details.

        OUTPUT:

        The closure of the normal cone. The `i`-th entry equals the
        value of the piecewise-linear function at the `i`-th point of
        the configuration.

        For an irregular triangulation, the normal cone is empty. In
        this case, a single point (the origin) is returned.

        EXAMPLES::

            sage: triangulation = polytopes.hypercube(2).triangulate(engine='internal')
            sage: triangulation
            (<0,1,3>, <1,2,3>)
            sage: N = triangulation.normal_cone();  N
            4-d cone in 4-d lattice
            sage: N.rays()
            ( 0,  0,  0, -1),
            ( 0,  0,  1,  1),
            ( 0,  0, -1, -1),
            ( 1,  0,  0,  1),
            (-1,  0,  0, -1),
            ( 0,  1,  0, -1),
            ( 0, -1,  0,  1)
            in Ambient free module of rank 4
            over the principal ideal domain Integer Ring
            sage: N.dual().rays()                                                       # needs sage.libs.pari
            (1, -1, 1, -1)
            in Ambient free module of rank 4
            over the principal ideal domain Integer Ring

        TESTS::

            sage: polytopes.simplex(2).triangulate().normal_cone()
            3-d cone in 3-d lattice
            sage: _.dual().is_trivial()                                                 # needs sage.libs.pari
            True
        """
        if not self.point_configuration().base_ring().is_subring(QQ):
            raise NotImplementedError('Only base rings ZZ and QQ are supported')
        from ppl import Constraint_System, Linear_Expression, C_Polyhedron
        from sage.matrix.constructor import matrix
        from sage.arith.functions import lcm
        pc = self.point_configuration()
        cs = Constraint_System()
        for facet in self.interior_facets():
            s0, s1 = self._boundary_simplex_dictionary()[facet]
            p = set(s0).difference(facet).pop()
            q = set(s1).difference(facet).pop()
            origin = pc.point(p).reduced_affine_vector()
            base_indices = [i for i in s0 if i != p]
            base = matrix([ pc.point(i).reduced_affine_vector()-origin for i in base_indices ])
            sol = base.solve_left( pc.point(q).reduced_affine_vector()-origin )
            relation = [0]*pc.n_points()
            relation[p] = sum(sol)-1
            relation[q] = 1
            for i, base_i in enumerate(base_indices):
                relation[base_i] = -sol[i]
            rel_denom = lcm([QQ(r).denominator() for r in relation])
            relation = [ ZZ(r*rel_denom) for r in relation ]
            ex = Linear_Expression(relation,0)
            cs.insert(ex >= 0)
        from sage.modules.free_module import FreeModule
        ambient = FreeModule(ZZ, self.point_configuration().n_points())
        if cs.empty():
            cone = C_Polyhedron(ambient.dimension(), 'universe')
        else:
            cone = C_Polyhedron(cs)
        from sage.geometry.cone import _Cone_from_PPL
        return _Cone_from_PPL(cone, lattice=ambient)

    def adjacency_graph(self):
        """
        Return a graph showing which simplices are adjacent in the triangulation.

        OUTPUT:

        A graph consisting of vertices referring to the simplices in the
        triangulation, and edges showing which simplices are adjacent to each
        other.

        .. SEEALSO::

            * To obtain the triangulation's 1-skeleton, use
              :meth:`SimplicialComplex.graph` through
              ``MyTriangulation.simplicial_complex().graph()``.

        AUTHORS:

        * Stephen Farley (2013-08-10): initial version

        EXAMPLES::

            sage: p = PointConfiguration([[1,0,0], [0,1,0], [0,0,1], [-1,0,1],
            ....:                         [1,0,-1], [-1,0,0], [0,-1,0], [0,0,-1]])
            sage: t = p.triangulate()
            sage: t.adjacency_graph()                                                   # needs sage.graphs
            Graph on 8 vertices
        """
        vertices = [Set(_) for _ in list(self)]
        from sage.graphs.graph import Graph
        return Graph([vertices,
                  lambda x,y: len(x-y) == 1])



# sage_setup: distribution = sagemath-polyhedra
r"""
Base class for mutable polyhedra.

Just like vectors and matrices they can be set immutable.
The constructor does this by default.
"""

from sage.misc.abstract_method import abstract_method

from .base import Polyhedron_base


class Polyhedron_mutable(Polyhedron_base):
    """
    Base class for polyhedra that allow mutability.

    This should not be used directly.
    """

    def __hash__(self):
        r"""
        TESTS::

            sage: p = Polyhedron([[1, 1]], mutable=True)
            sage: set([p])
            Traceback (most recent call last):
            ...
            TypeError: mutable polyhedra are unhashable
            sage: p.set_immutable()
            sage: set([p])
            {A 0-dimensional polyhedron in ZZ^2 defined as the convex hull of 1 vertex}
        """
        if self._is_mutable:
            raise TypeError("mutable polyhedra are unhashable")
        return Polyhedron_base.__hash__(self)

    def _clear_cache(self):
        r"""
        Clear the Vrepresentation and Hrepresentation data of ``self``.

        TESTS::

            sage: p = polytopes.permutahedron(4)
            sage: P = p.parent()
            sage: q = P._element_constructor_(p, mutable=True)
            sage: TestSuite(q).run()                                                    # needs sage.libs.pari
            sage: q._clear_cache()
            sage: TestSuite(q).run()                                                    # needs sage.libs.pari

        ::

            sage: q.set_immutable()
            sage: q._clear_cache()
            Traceback (most recent call last):
            ...
            TypeError: cannot clear cache of immutable polyhedra
        """
        if not self._is_mutable:
            raise TypeError("cannot clear cache of immutable polyhedra")

        # Invalidate object pointing towards this polyhedron (faces etc.).
        for ob in self._dependent_objects:
            ob._polyhedron = None
        backend_object = self.__dict__["_" + self._backend_object_name]
        del self.__dict__
        self.__dict__["_" + self._backend_object_name] = backend_object
        self._is_mutable = True
        self._dependent_objects = []

    def _add_dependent_object(self, ob):
        r"""
        Add an object that has ``self`` has attribute ``_polyhedron``.

        When ``self`` is modified, we delete this attribute to invalidate those objects.

        EXAMPLES::

            sage: p = Polyhedron([[1, 1]], mutable=True)
            sage: class foo:
            ....:     def __init__(self, p):
            ....:         self._polyhedron = p
            ....:
            sage: a = foo(p)
            sage: a.__dict__
            {'_polyhedron': A 0-dimensional polyhedron in ZZ^2 defined as the convex hull of 1 vertex}
            sage: p._add_dependent_object(a)
            sage: p._clear_cache()
            sage: a.__dict__
            {'_polyhedron': None}

        TESTS::

            sage: from sage.geometry.newton_polygon import NewtonPolygon
            sage: p = Polyhedron([[1, 1]], mutable=True)
            sage: n = NewtonPolygon(p)
            sage: n
            Finite Newton polygon with 1 vertex: (1, 1)
            sage: n = NewtonPolygon(p)
            sage: p._clear_cache()
            sage: n
            <repr(<sage.geometry.newton_polygon.ParentNewtonPolygon_with_category.element_class at ...>) failed: AttributeError: 'NoneType' object has no attribute 'vertices'>

        ::

            sage: f = p.faces(0)[0]; f
            A 0-dimensional face of a Polyhedron in ZZ^2 defined as the convex hull of 1 vertex
            sage: p._clear_cache()
            sage: f
            <repr(<sage.geometry.polyhedron.face.PolyhedronFace at ...>) failed: AttributeError: 'NoneType' object has no attribute 'parent'>

        ::

            sage: v = p.vertices()[0]
            sage: p = Polyhedron([[1, 1]], mutable=True)
            sage: v = p.Vrepresentation(0); v
            A vertex at (1, 1)
            sage: h = p.Hrepresentation(0); h
            An equation (0, 1) x - 1 == 0
            sage: p._clear_cache()
            sage: v.polyhedron() is None
            True
            sage: h.polyhedron() is None
            True

        ::

            sage: p = Polyhedron([[1, 0], [0, 1]], mutable=True)
            sage: r = p.relative_interior()
            sage: p._clear_cache()
            sage: r
            Relative interior of None
        """
        if ob._polyhedron is not self:
            raise ValueError
        self._dependent_objects.append(ob)

    def is_mutable(self):
        r"""
        Return ``True`` if the polyhedron is mutable, i.e. it can be modified in place.

        EXAMPLES::

            sage: p = Polyhedron([[1, 1]], mutable=True)
            sage: p.is_mutable()
            True
            sage: p = Polyhedron([[1, 1]], mutable=False)
            sage: p.is_mutable()
            False
        """
        return self._is_mutable

    def is_immutable(self):
        r"""
        Return ``True`` if the polyhedron is immutable, i.e. it cannot be modified in place.

        EXAMPLES::

            sage: p = Polyhedron([[1, 1]], mutable=True)
            sage: p.is_immutable()
            False
            sage: p = Polyhedron([[1, 1]], mutable=False)
            sage: p.is_immutable()
            True
        """
        return not self._is_mutable

    @abstract_method
    def set_immutable(self):
        r"""
        Make this polyhedron immutable. This operation cannot be undone.

        TESTS::

            sage: from sage.geometry.polyhedron.base_mutable import Polyhedron_mutable
            sage: p = polytopes.cube()
            sage: Polyhedron_mutable.set_immutable(p)
            Traceback (most recent call last):
            ...
            TypeError: 'AbstractMethod' object is not callable
        """

    @abstract_method
    def Vrepresentation(self):
        r"""
        A derived class must overwrite such that it restores Vrepresentation
        after clearing it.

        TESTS::

            sage: from sage.geometry.polyhedron.base_mutable import Polyhedron_mutable
            sage: p = polytopes.cube()
            sage: Polyhedron_mutable.Vrepresentation(p)
            Traceback (most recent call last):
            ...
            TypeError: 'AbstractMethod' object is not callable
        """
        # A derived class must implemented it to recalculate, if necessary.

    @abstract_method
    def Hrepresentation(self):
        r"""
        A derived class must overwrite such that it restores Hrepresentation
        after clearing it.

        TESTS::

            sage: from sage.geometry.polyhedron.base_mutable import Polyhedron_mutable
            sage: p = polytopes.cube()
            sage: Polyhedron_mutable.Hrepresentation(p)
            Traceback (most recent call last):
            ...
            TypeError: 'AbstractMethod' object is not callable
        """
        # A derived class must implemented it to recalculate, if necessary.
